# SPDX-FileCopyrightText: 2021 G2Elab / MAGE
#
# SPDX-License-Identifier: Apache-2.0

from scipy.stats import qmc
from jax import jacfwd,jit,config
from noloadj.optimization.Tools import StructList
from typing import Callable, Dict, List, Any
from pandas import DataFrame as df
import numpy as np,sys
import jax.numpy as jnp
config.update("jax_enable_x64", True)

'''Derivative-based Global Sensitivity Measure'''

def computeDGSM(model: Callable[..., Dict], inputs: Dict[str, Any],
                outputs: List[str], deltas:Dict[str,Any],
                deltas_type:Dict[str,Any],N:int,Param: Dict[str, Any]={}):
    '''
    Compute DGSM (Derivative-based Global Sensitivity Measure) of a model.

    :param model: model
    :param inputs: dict: names and initial values of derivable input variables
    :param outputs: list: names of output variables
    :param deltas: dict: variation of derivable input variables
    :param deltas_type: dict: type of variation for each derivable input ('rel' for relative or 'abs' for absolute)
    :param N: int: number of points to compute DGSM. Must be a power of 2.
    :param Param: names and values of non-derivable input variables (dictionary)
    :return: dataframe : DGSM
    '''
    if int(bin(N)[3:])!=0: # test si N est une puissance de 2
        print('Warning :', N, 'is not a power of 2 (512, 1024, 2058, ...)')
        sys.exit(0)

    m = int(np.log(N) / np.log(2.)) # 2^(m)=N

    inputs_names,init_values=list(inputs.keys()),list(inputs.values())
    len_inputs,len_outputs=len(inputs_names),len(outputs)

    vi=np.empty((len_inputs,len_outputs))
    var=np.empty((len_inputs,len_outputs))
    b_a={}

    sampler = qmc.Sobol(d=len_inputs, scramble=False) 
    sample = sampler.random_base2(m=m) # création des échantillons [0,1]
    l_bounds,u_bounds=[],[]

    for input in inputs_names:
        index_input=inputs_names.index(input)
        if deltas_type[input]=='abs':
            l_bounds.append(init_values[index_input]-deltas[input])
            u_bounds.append(init_values[index_input]+deltas[input])
            b_a[input]=2.*deltas[input]
        elif deltas_type[input]=='rel':
            l_bounds.append(init_values[index_input]-deltas[input]*
                            init_values[index_input])
            u_bounds.append(init_values[index_input]+deltas[input]*
                            init_values[index_input])
            b_a[input]=init_values[index_input]*((1.+deltas[input])-(
                    1.-deltas[input]))

    scale_samples=qmc.scale(sample, l_bounds,u_bounds).T # mise à échelle des échantillons

    for input in inputs_names:
        index_input=inputs_names.index(input)
        inputs[input] = scale_samples[index_input]
                                  
    outputs_values = computeParametricValue(model,inputs,outputs,N,Param) # évaluation modèle N fois
    variances=np.var(outputs_values,axis=1) # calcul des variances de chaque sortie
    
    jac_values=computeParametricJacobian(model,inputs,outputs,N,Param) # gradients modèle N fois

    for input in inputs_names:
        index_input=inputs_names.index(input)

        for j in range(len_outputs):
            vi[index_input][j]=np.mean(jac_values[j*len_inputs+index_input]**2)
            var[index_input][j]=b_a[input]*b_a[input]/(np.pi*np.pi*variances[j])

    dgsm=df(vi*var,index=inputs_names,columns=outputs)

    return dgsm


def computeParametricValue(model:Callable[...,Dict],inputs_range: Dict[str,Any],
                      outputs: List[str],N:int,Param: Dict[str, Any]={}):
                      
    outputs_values=np.empty((N,len(outputs)))
    inputs={}
    
    for i in range(N):
        for key in list(inputs_range.keys()):
            inputs[key]=inputs_range[key][i]
        if Param!={}:
            res = model(**inputs,**Param)
        else:
            res = model(**inputs)
        dico = {k: v for k, v in res.__iter__()}  # conversion en dictionnaire
        out=[]
        nans_vars = []
        for vars in outputs:
            try:
                if vars not in list(dico.keys()):
                    raise KeyError(vars) #si la variable du cahier des charges
                if jnp.all(jnp.isnan(dico[vars])):
                    raise ValueError(vars)
            except KeyError: # n'appartient pas aux sorties du modèle
                print('Warning :',vars,'is not in model')
                sys.exit(0)
            except ValueError:
                nans_vars.append(vars)
            else:
                out.append(dico[vars])
        if nans_vars != []:
            print('Warning :', nans_vars, 'is Nan')
            sys.exit(0)
        outputs_values[i]=out

    return np.transpose(outputs_values)


def computeParametricJacobian(model:Callable[...,Dict], inputs_range:
            Dict[str,Any],outputs: List[str],N:int,Param: Dict[str, Any]={}):

    outputsjac_values=np.empty((N,len(outputs)*len(inputs_range)))
    inputs={}

    def eval_model(array_inputs):
        if Param!={}:
            res = model(*array_inputs,**Param)
        else:
            res = model(*array_inputs)
        dico = {k: v for k, v in res.__iter__()}  # conversion en dictionnaire
        out = []
        nans_vars = []
        for vars in outputs:
            try:
                if vars not in list(dico.keys()):
                    raise KeyError(vars)  # si la variable du cahier des charges
                #if jnp.all(jnp.isnan(dico[vars])):
                #    raise ValueError(vars)
            except KeyError:  # n'appartient pas aux sorties du modèle
                print('Warning :', vars, 'is not in model')
                sys.exit(0)
            #except ValueError:
            #    nans_vars.append(vars)
            else:
                out.append(dico[vars])
        if nans_vars != []:
            print('Warning :', nans_vars, 'is Nan')
            sys.exit(0)
        return out

    eval_model_jit=jit(eval_model)
    
    for i in range(N):
        for key in list(inputs_range.keys()):
            inputs[key]=inputs_range[key][i]

        array_inputs=jnp.array(list(inputs.values()))
        out=jacfwd(eval_model_jit)(array_inputs)
        outputsjac_values[i]=Flatten(out)

    return np.transpose(outputsjac_values)

def Flatten(vector):
    res=[]
    for i in range(len(vector)):
        for j in range(len(vector[i])):
            res.append(vector[i][j])
    return res


# SPDX-FileCopyrightText: 2021 G2Elab / MAGE
#
# SPDX-License-Identifier: Apache-2.0

from noloadj.analyse.Sobol import computeSobol
from noloadj.analyse.simulation import computeOnce,computeJacobian,\
    computeParametric
from noloadj.analyse.DGSM import computeDGSM
import numpy as np
import jax.numpy as jnp

def rosenbrock(x,y):
    fobj=(1-x)*(1-x)+100*(y-x*x)**2
    ctr1=(x-1)**3-y+1
    ctr2=x+y-2
    return locals().items()

def Ishigami(x1,x2,x3,A,B):
    Y=jnp.sin(x1)+A*jnp.sin(x2)*jnp.sin(x2)+B*jnp.sin(x1)*(x3**4)
    return locals().items()


# Compute outputs
inputs = {'x': 1.0, 'y': 2.0}
outputs = ['fobj', 'ctr1', 'ctr2']
results = computeOnce(model=rosenbrock, inputs=inputs, outputs=outputs)
print(outputs, '=', results)


# Compute Parametric
inputs = {'y': 2.0}
outputs = ['fobj', 'ctr1', 'ctr2']

variable = 'x'
values = np.arange(-1.5, 1.5, 0.1)  # [-1.5, -1.4, ..., 1.5]
iter = computeParametric(rosenbrock, variable, values, inputs, outputs)

df = iter.print()
print(df)
iter.plotXY()

# Compute Gradients
inputs = {'x': 1.0, 'y': 2.0}
outputs = ['fobj', 'ctr1', 'ctr2']
dfobj,dctr1,dctr2 = computeJacobian(model=rosenbrock, inputs=inputs,
                                    outputs=outputs)
print('dfobj =', dfobj)
print('dctr1 =', dctr1)
print('dctr2 =', dctr2)

# Compute DGSM
inputs = {'x1': 0., 'x2': 0., 'x3': 0.}
deltas = {'x1': jnp.pi, 'x2': jnp.pi, 'x3': jnp.pi}
deltas_type = {'x1': 'abs', 'x2': 'abs', 'x3': 'abs'}
outputs = ['Y']
Param = {'A': 7., 'B': 0.1}
DGSM = computeDGSM(model=Ishigami, inputs=inputs, outputs=outputs,
                   deltas=deltas, deltas_type=deltas_type, N=1024,Param=Param)
print('DGSM=', DGSM)

# Compute Sobol 1st-order indices
inputs = {'x1': 0., 'x2': 0., 'x3': 0.}
deltas = {'x1': jnp.pi, 'x2': jnp.pi, 'x3': jnp.pi}
deltas_type = {'x1': 'abs', 'x2': 'abs', 'x3': 'abs'}
outputs = ['Y']
Param = {'A': 7., 'B': 0.1}
Sobol = computeSobol(model=Ishigami, inputs=inputs, outputs=outputs,
                   deltas=deltas, deltas_type=deltas_type, N=1024,Param=Param)
print('Sobol=',Sobol)




# SPDX-FileCopyrightText: 2021 G2Elab / MAGE
#
# SPDX-License-Identifier: Apache-2.0
"""
The aim of these Paretotools functions is to find the longest distance on
axes x or y for every couple of points that can solve the optimization problem.
"""
import numpy as np
def getXorYmid(xy, normX, normY, xmin, ymin):
    """
    Finds the middle point between x and y.

    :param xy: list of tuples including x and y
    :param normX: float: norm of X
    :param normY: float:norm of Y
    :param xmin: float:minimial abscisse of the xy couple
    :param ymin: float:minimial ordinate of the xy couple
    :return: the index of the objective function (int), the middle of [XY] (float).
    """
    xyT=np.array(xy).T
    xyNorm=np.array([((xyT[0,:]-xmin)/normX),((xyT[1,:]-ymin)/normY)]).T.\
        tolist()

    #p1, p2, d = maxDist(xyNorm)
    p1, p2 = maxDist(xyNorm)
    dX = abs(p2[0]-p1[0])    #on suppose que p2 est bien plus grand que p1
    dY = abs(p2[1]-p1[1])
    if dX > dY :
        return 0, (p1[0]+p2[0])/2*normX+xmin
    else:
        return 1, (p1[1]+p2[1])/2*normY+ymin


def maxDist(points):
    points = sorted(points, key=lambda x: x[0])  # Presorting x-wise
    pts2 = [pt for pt in points] #copy
    pts1 = np.array(points[0:-1]).T
    pts2 = np.array(pts2[1:]).T

    dist = (pts1[0] - pts2[0]) ** 2 + (pts1[1] - pts2[1]) ** 2
    index = np.argmax(dist)
    return points[index], points[index+1]



if __name__ == '__main__':
    a = [[9953.86318652108,3702.5292382627013],
    [16922.450708122255,1312.920633183176],
    [13483.690450357713,1790.842361273781],
    [11890.18707155824,2268.764075340589],
    [10963.413555536845,2746.685796235864],
    [10358.889713185245,3224.6075172528217],
    [11347.58068777548,2520.915738089411],
    [12741.29819530203,1977.522534854665],
    [14135.015696837956,1661.5621851434228],
    [15528.733518913752,1456.2150724085136]]

    x=[xy[0] for xy in a]
    y=[xy[1] for xy in a]

    #print( maxDist(a) )
    #print( getXorYmid(a, 15528-9953, 3702-1456, 9953,1456) )
    print(maxDist(a))




import jax.numpy as np
from jax.lax import *
from jax import custom_jvp,jvp
from functools import partial
from jax import config
config.update("jax_enable_x64", True)

def odeint44(f,x0,vect_t,*P):
    '''
    Solves an ODE system described by f with Runge-Kutta 44 algorithm.

    :param f: a class that describes the ODE system
    :param x0: initial state vector
    :param vect_t: time vector on which the ODE system must be solved
    :param P: optimization inputs
    :return: Several outputs :

    - state vector 'xf' values across time vector
    - output vector 'yf' values across time vector
    '''
    return _odeint44(f,x0,vect_t,*P)


def rk_step(x_prev, t_prev, h,f,*P):
    '''
    An iteration of the Runge-Kutta 44 algorithm.

    :param x_prev: the state vector at the previous iteration
    :param t_prev: the time at the previous iteration
    :param h: constant step size
    :param f: a class that describes the ODE system
    :param P: optimization inputs
    :return: the state vector x_now and time t_now at the present iteration.
    '''
    k1 = f.timederivatives(x_prev, t_prev,*P)
    k2 = f.timederivatives(x_prev + h*0.5 * k1, t_prev + 0.5 * h,*P)
    k3 = f.timederivatives(x_prev + h*0.5 * k2, t_prev + 0.5 * h,*P)
    k4 = f.timederivatives(x_prev + h*k3, t_prev + h,*P)

    x_now = x_prev + h *(k1 + 2 * k2 + 2 * k3 + k4) / 6
    t_now = t_prev + h
    return x_now,t_now


def next_step_simulation(x_prev,t_prev,h,f,*P):
    '''
    Computes the state vector, output vector and time at the present iteration.

    :param x_prev: the state vector at the previous iteration
    :param t_prev: the time at the previous iteration
    :param h: constant step size
    :param f: a class that describes the ODE system
    :param P: optimization inputs
    :return: the state vector x_now, outputs vector y_now and time t_now at
        the present iteration.
    '''
    x_now,t_now=rk_step(x_prev,t_prev,h,f,*P)
    y_now=f.output(x_now,t_now,*P)
    return x_now,t_now,y_now

@partial(custom_jvp,nondiff_argnums=(0,))
def _odeint44(f,x0,vect_t,*P):
    h=vect_t[1]-vect_t[0] # constant step size

    def scan_fun(state,te): # iteration of simulation

        x_prev,t_prev,y_prev=state

        x_now,t_now,y_now=next_step_simulation(x_prev,t_prev,h,f,*P)

        return (x_now,t_now,y_now),(x_now,y_now)


    y0=f.output(x0,vect_t[0],*P)
    vect,(xf,yf)=scan(scan_fun,(x0,vect_t[0],y0),vect_t[1:])

    xf=np.transpose(np.concatenate((x0[None], xf)))
    yf=np.transpose(np.concatenate((y0[None], yf)))

    return xf,yf


@_odeint44.defjvp
def _odeint44_jvp(f, primals, tangents):
    '''
    Solves an ODE system described by df/dP with Runge-Kutta 44 algorithm.

    :param f: a class that describes the ODE system
    :param primals: tuple including initial state vector, time vector and
     optimization inputs.
    :param tangents: tuple including differentials of initial state vector,
        time vector and optimization inputs.
    :return: a matrix xf with state vector values across time, and a matrix yf
        with output vector values across time, and their respective derivatives
        w.r.t P (dxf and dyf).
    '''
    x0, vect_t, *P = primals
    dx0, _, *dP = tangents
    nPdP = len(P)

    xf,dxf,yf,dyf = odeint44_etendu(f,nPdP,x0,dx0,vect_t,*P,*dP)
    return (xf,yf),(dxf,dyf)

def f_grads(x,dx, t, f,nPdP,*P_and_dP):
    '''
    Computes the derivatives of state vector w.r.t P.

    :param x: state vector values
    :param dx: differential of state vector
    :param t: time at present iteration
    :param f: a class that describes the ODE system
    :param nPdP: length of optimization inputs
    :param P_and_dP: optimization inputs values and differentials
    :return: the derivatives of state vector w.r.t P.
    '''
    P, dP = P_and_dP[:nPdP], P_and_dP[nPdP:]
    res, dres = jvp(f.timederivatives, (x, t, *P), (dx, 0., *dP))
    return dres

def rk44_step_der(x_prev, t_prev, dx_prev,h,f,nPdP,*P_and_dP):
    '''
    An iteration of the Runge-Kutta 44 algorithm for the ODE system described
    by df/dP.

    :param x_prev: the state vector at the previous iteration
    :param t_prev: the time at the previous iteration
    :param dx_prev: the differential of state vector at the previous iteration
    :param h: constant step size
    :param f: a class that describes the ODE system.
    :param nPdP: length of optimization inputs
    :param P_and_dP: optimization inputs values and differentials
    :return: dx/dp and x at the present iteration.
    '''
    P,_ = P_and_dP[:nPdP],P_and_dP[nPdP:]
    k1 = f.timederivatives(x_prev, t_prev,*P)
    k2 = f.timederivatives(x_prev + h*0.5 * k1, t_prev + 0.5 * h,*P)
    k3 = f.timederivatives(x_prev + h*0.5 * k2, t_prev + 0.5 * h,*P)
    k4 = f.timederivatives(x_prev + h*k3, t_prev + h,*P)

    dk1 = f_grads(x_prev, dx_prev, t_prev,f,nPdP, *P_and_dP)
    dk2 = f_grads(x_prev + h*0.5 * k1, dx_prev + h * 0.5 * dk1,t_prev +
                0.5 * h,f,nPdP,*P_and_dP)
    dk3 = f_grads(x_prev + h*0.5 * k2, dx_prev + h * 0.5 * dk2,t_prev +
                0.5 * h,f,nPdP,*P_and_dP)
    dk4 = f_grads(x_prev + h * k3,dx_prev + h * dk3,t_prev+h,f,nPdP,*P_and_dP)

    x_now = x_prev + h * (k1 + 2 * k2 + 2 * k3 + k4) / 6
    dx_now = dx_prev + h *(dk1 + 2 * dk2 + 2 * dk3 + dk4) / 6
    return dx_now,x_now

def next_der_step_simulation(x_prev,t_prev,dx_prev,h,f,
                             nPdP,*P_and_dP):
    '''
    Computes the state vector, output vector, their derivatives and time at
    the present iteration.

    :param x_prev: the state vector at the previous iteration
    :param t_prev: the time at the previous iteration
    :param dx_prev: the differentials of the state vector at the previous iteration
    :param h: constant step size
    :param f: a class that describes the ODE system
    :param nPdP: length of optimization inputs
    :param P_and_dP: optimization inputs values and differentials
    :return: the state vector x_now, outputs vector y_now, their derivatives
     dx_now, dy_now, and time t_now at the present iteration.
    '''
    P,dP = P_and_dP[:nPdP],P_and_dP[nPdP:]
    dx_now,x_now = rk44_step_der(x_prev, t_prev, dx_prev,h,f,nPdP,
                              *P_and_dP)
    t_now=t_prev+h
    y_now=f.output(x_now,t_now,*P)
    dy_now = jvp(f.output, (x_now, t_now, *P), (dx_now, 0., *dP))[1]
    return dx_now,dy_now,x_now,y_now,t_now

def odeint44_etendu(f,nPdP,x0,dx0,vect_t,*P_and_dP):
    P,dP = P_and_dP[:nPdP],P_and_dP[nPdP:]
    h=vect_t[1]-vect_t[0] # constant step size

    def scan_fun(state, te): # iteration of simulation

        x_prev,dx_prev,y_prev,dy_prev,t_prev=state

        dx_now,dy_now,x_now,y_now,t_now=next_der_step_simulation(x_prev,t_prev,
                            dx_prev, h,f,nPdP,*P_and_dP)

        return (x_now,dx_now,y_now,dy_now,t_now), (x_now,dx_now,y_now,dy_now)

    for element in f.__dict__.keys(): # pour eviter erreurs de code
        if hasattr(f.__dict__[element],'primal'):
            f.__dict__[element]=f.__dict__[element].primal

    y0,dy0=jvp(f.output,(x0,vect_t[0],*P),(dx0,0.,*dP))
    vect,(xf,dxf,yf,dyf)=scan(scan_fun,(x0,dx0,y0,dy0,
                        vect_t[0]),vect_t[1:])

    xf=np.transpose(np.concatenate((x0[None], xf)))
    yf=np.transpose(np.concatenate((y0[None], yf)))
    dxf=np.transpose(np.concatenate((dx0[None], dxf)))
    dyf = np.transpose(np.concatenate((dy0[None], dyf)))

    return xf,dxf,yf,dyf




# SPDX-FileCopyrightText: 2021 G2Elab / MAGE
#
# SPDX-License-Identifier: Apache-2.0
import jax.numpy as jnp
import numpy as np
import jax
from jax import custom_jvp
from functools import partial

'''To deal with constraints'''
class StructList:
    """
    Attributes :

    - List : list of all scalar/vector constraints
    - shape = list containing the sizes of the vectors of the different constraints (0 if scalar, from 1 if vector)
    - etat = string indicating if the list has already been flattened or not (unflattened)

    When creating an object of the class, if the state of the list is 'unflattened', the shape vector is calculated automatically. Otherwise it must be provided when creating the object.
    """
    # List = None # liste regroupant l'ensemble des contraintes scalaires et/ou
    # # vectorielles
    # shape = [] # liste regroupant les tailles des vecteurs des différentes
    # # contraintes (0 si scalaire, à partir de 1 si vectorielle)
    # etat = '' # chaîne de caractères indiquant si la liste est déjà aplatie
    # # (flattened) ou non (unflattened)

    def __init__(self, List=[], etat='unflattened', shape=[]):
        sh = []
        self.List = List
        self.etat = etat
        if self.etat == 'unflattened': # si la liste n'est pas "aplatie"
            if len(self.List) == 1 and isinstance(self.List[0], (int, float)):
                # cas : 1 contrainte scalaire
                shape = [0]
            else:
                for i in range(len(self.List)): # on parcourt les éléments
                    # de la liste des contraintes
                    if isinstance(self.List[i], (list, np.ndarray)): # si la
                        # contrainte est vectorielle
                        sh.append(len(self.List[i])) # on ajoute à la liste
                        # shape la longueur du vecteur de contraintes
                    else:   # si la contrainte est scalaire
                        sh.append(0) # on ajoute un 0
                        # (signe que la contrainte est scalaire)
                shape = sh
        self.shape = shape

    def flatten(self):
        """
        "Flattens" a list of mixed constraints (from 'unflattened' to 'flattened').

        :return: the "flattened" list
        """
        if len(self.List) == 1 and isinstance(self.List[0], (int, float)):
            # cas s: 1 contrainte scalaire
            return self.List
        res = []
        for i in range(len(self.List)): # on parcourt chaque type de
            # contrainte dans la liste
            if isinstance(self.List[i], (list, np.ndarray)): # si la
                # contrainte est vectorielle
                for j in range(len(self.List[i])): # pour chaque composante
                    # de cette contrainte vectorielle
                    res.append(self.List[i][j]) # on ajoute chaque composante
                    # au résultat final res
            else: # si la contrainte est scalaire
                res.append(self.List[i]) # on l'ajoute au résultat final res
        self.etat = 'flattened' # on change l'attribut état de la liste pour
        # signifier qu'elle a été aplatie
        self.List=res
        return res # on renvoie la liste aplatie

    def unflatten(self): # fonction pour "reformer" une liste de contraintes
        # complexes "aplatie"
        """
        "Reforms" a "flattened" list of mixed constraints (from 'flattened' to
        'unflattened').

        :return: the "reformed" list
        """
        res = []
        i = 0
        for size in self.shape: # on parcourt les éléments de "shape"
            if size == 0:  #si la contrainte est scalaire
                res.append(self.List[i])
                i = i + 1
            else: # si la contrainte est vectorielle
                res.append(self.List[i:i + size])
                i = i + size
        self.etat = 'unflattened' # on change l'attribut état de la liste pour
        # signifier qu'elle a été "reformée"
        self.List=res
        return res

def normalize(x,bounds):
    '''
    Normalizes optimization variables between 0 and 1.

    :param x: variables to optimize
    :param bounds: bounds for variables
    :return: normalized values between 0 and 1
    '''
    Tbounds = np.transpose(bounds)
    Diffbounds = Tbounds[1] - Tbounds[0]
    x=(x-Tbounds[0])/Diffbounds
    return x


def denormalize(x,bounds):
    '''
    Denormalizes the normalized variables.

    :param x: normalized variables
    :param bounds: bounds for variables
    :return: original values of optimization variables
    '''
    Tbounds=np.transpose(bounds)
    Diffbounds=Tbounds[1]-Tbounds[0]
    x=x*Diffbounds+Tbounds[0]
    return x


def normalize_bounds(bounds):
    '''
    Normalizes bounds of optimization variables between 0 and 1.

    :param bounds: original bounds of variables
    :return: normalized bounds between 0 and 1
    '''
    norm_bounds=[[0,1] for i in range(len(bounds))]
    return norm_bounds


###################################################### external functions
def compute_external_function(inputs,Extfunc,len_output):
    '''
    Executes external function in optimization loops. Gives the output of the external fucntion.

    :param inputs: JaxArray: inputs of the external function.
    :param Extfunc: An instance of the class regrouping external function.
    :param len_output: int : len of the output vector of the external function
    :return: JaxArray : output vector
    '''
    return _compute_external(inputs,len_output,Extfunc)
@partial(custom_jvp,nondiff_argnums=(1,2))
def _compute_external(inputs,len_output,Extfunc):
    output_type=jax.ShapeDtypeStruct((len_output,),'float64')
    def compute_intermediary_function(inputs):
        outputs=Extfunc.compute(inputs)
        return np.array(outputs)
    outputs=jax.pure_callback(compute_intermediary_function,output_type,inputs)
    return outputs

def jacobien_compute_external(len_output,Extfunc,primals,tangents):
    '''
    Executes external functions in optimization loops. Gives the derivative of output w.r.t inputs.

    :param len_output: int : len of the output vector of the external function
    :param Extfunc: An instance of the class regrouping external function.
    :param primals: JaxArray: inputs of the external function.
    :param tangents: JaxArray: differential inputs of the external function.
    :return: JaxArray : derivative of output vector w.r.t inputs.
    '''
    inputs,=primals
    dinputs,=tangents
    output_type=jax.ShapeDtypeStruct((len_output,),'float64')
    def compute_intermediary_function(inputs):
        outputs=Extfunc.compute(inputs)
        return np.array(outputs)
    def jacobian_intermediary_function(inputs):
        gradients=Extfunc.jacobian(inputs)
        return np.array(gradients)
    outputs=jax.pure_callback(compute_intermediary_function,output_type,inputs)
    gradient_type=jax.ShapeDtypeStruct((len_output,len(inputs)),'float64')
    gradients=jax.pure_callback(jacobian_intermediary_function,gradient_type,
                                inputs)
    return outputs,jax.numpy.dot(gradients,dinputs)

_compute_external.defjvp(jacobien_compute_external)
#####################################################

def Print(elements):
    from jax import debug
    debug.print('{}',elements)


if __name__ == "__main__":
    seq = [1, [1], [2,3], 4, [5, 6, 7], np.array([1, 2])] #, [[1], 2, [3, [4]]]]
    print("original:\t", seq)
    seq2 = StructList(seq)
    res = seq2.flatten()
    print("flattened:\t", res)
    print("shape:\t\t", seq2.shape)
    seq3 = StructList(res,'flattened',seq2.shape)
    res2 = seq3.unflatten()
    print("unflattened:", res2)



import tkinter as tk
from tkinter import ttk,TOP,BOTH
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import sys,pandas
import numpy as np
from openpyxl.styles import Font, Color
from openpyxl import Workbook
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg,\
    NavigationToolbar2Tk
### Tk App ###

class App:
    '''
    Class to create Graphical User Interface.

    Attributes :

    - root: the interface with a Frame.
    - options: list including the names of optimization variables.
    - df: dataframe including the optimization variables values across iterations.
    - specifications: dataframe presenting the specifications of the optimization problem.

    '''
    def __init__(self, root, options, df, specifications):
        '''
        Initialization of the class App
        :param root: the interface with a Frame
        :param options: list including the names of optimization variables.
        :param df: dataframe including the optimization variables values across iterations.
        :param specifications: dataframe including the specifications of the optimization problem.
        '''
        self.root = root
        self.options = options
        self.selected_options = []  # Keep track of selected options
        self.df = df
        self.specifications = specifications
        self.selected_x = tk.StringVar()
        self.selected_y = tk.StringVar()

        self.create_widgets()

    def create_widgets(self):

        # Create a Treeview widget
        self.treeview = ttk.Treeview(self.root, columns=("Variable", "Min",
                    "Max", "Value", "Type", "Input/Output"), show="headings",
                            height=10)
        self.treeview.pack(padx=10, pady=5)

        # Add column headings
        self.treeview.heading("Variable", text="Variable")
        self.treeview.heading("Min", text="Min")
        self.treeview.heading("Max", text="Max")
        self.treeview.heading("Value", text="Value")
        self.treeview.heading("Type", text="Type")
        self.treeview.heading("Input/Output", text="Input/Output")

        # Define color tags based on type and min/max columns
        type_colors = {
            #"bounds": "black",
            "ineq_cstr": "grey",
            "objective": "red",
            "eq_cstr": "green",
            #"free": "black",
        }

        type_names = {
            "bounds": "constrained",
            "ineq_cstr": "inequality",
            "objective": "objective",
            "eq_cstr": "fixed",
            "free": "free",
        }

        # Add options and additional information
        for option in self.options:
            type_ = self.specifications.Type[option]
            value = self.specifications.Value[option]
            input_output = self.specifications.In_Out[option]

            if type_=='bounds' or type_=='ineq_cstr':
                if isinstance(value[0],list):
                    min_value, max_value=str(np.transpose(np.array(value))[0]),\
                                          str(np.transpose(np.array(value))[1])
                else:
                    min_value, max_value = value
                value_text = ""

            elif type_=='eq_cstr':
                min_value = ""
                max_value = ""
                value_text = str(value)

            else:
                min_value = ""
                max_value = ""
                value_text = ""


            # Add rows and colors
            self.treeview.insert("", "end", values=(option, min_value,
               max_value, value_text, type_names.get(type_), input_output),
                                 tags=(type_,))
            self.treeview.tag_configure(type_, foreground=
            type_colors.get(type_, "black"))

        # Create a scrollbar for the Treeview
        scrollbar = ttk.Scrollbar(self.root, orient="vertical",
                                  command=self.treeview.yview)
        self.treeview.configure(yscrollcommand=scrollbar.set)
        scrollbar.pack(side="right", fill="y")

        # Create a context menu for the Treeview
        self.option_menu = tk.Menu(self.root, tearoff=False)
        self.option_menu.add_command(label="Additional Info",
                                     command=self.show_additional_info)
        self.option_menu.entryconfigure(0, state="disabled")  # Disable the "Additional Info" menu item initially
        self.treeview.bind("<Button-3>", self.show_context_menu)

        self.option_frame = ttk.Frame(self.root)
        self.option_frame.pack(padx=10, pady=10)

        self.x_label = ttk.Label(self.option_frame, text="X:")
        self.x_label.grid(row=0, column=0, padx=5, pady=5, sticky=tk.E)

        self.x_combobox = ttk.Combobox(self.option_frame, textvariable=
            self.selected_x, values=self.options)
        self.x_combobox.grid(row=0, column=1, padx=5, pady=5)

        self.y_label = ttk.Label(self.option_frame, text="Y:")
        self.y_label.grid(row=1, column=0, padx=5, pady=5, sticky=tk.E)

        self.y_combobox = ttk.Combobox(self.option_frame, textvariable=
            self.selected_y, values=self.options)
        self.y_combobox.grid(row=1, column=1, padx=5, pady=5)

        self.plot_button = ttk.Button(self.option_frame,
                    text="Plot Y / X", command=self.plot_sorted_X)
        self.plot_button.grid(row=3, column=0, columnspan=2, padx=5, pady=10)

        self.export_button = tk.Button(self.root, text="Export",
                                       command=self.export_to_excel)
        self.export_button.pack(pady=5)
        self.canvas=None
        self.toolbar=None
        self.exit_button = tk.Button(self.root, text="Exit",
                                     command=self.exit_program)
        self.exit_button.pack(pady=5)

    def show_context_menu(self, event):
        selection = self.treeview.selection()
        if selection:
            self.selected_options = [self.options[self.treeview.index(item)]
                                     for item in selection]  # Update selected options
            menu_label = f"Additional Info ({len(self.selected_options)} " \
                         f"options selected)"
            self.option_menu.entryconfigure(0, label=menu_label)
            self.option_menu.entryconfigure(0, state="normal")  # Enable the "Additional Info" menu item

            self.option_menu.entryconfigure(1, label="Plot",
                                    command=self.plot_multiple_selected_options)

            self.option_menu.tk_popup(event.x_root, event.y_root)

    def show_additional_info(self):
        for selected_option in self.selected_options:
            print(f"Additional info for {selected_option}")


    def plot_multiple_selected_options(self):
        '''
        Plots multiple selected variables in several graphs.

        :return: /
        '''
        for selected_option in self.selected_options:
            self.plot_one_selected_option(selected_option)

    def export_to_excel(self):
        '''
        Exports the results under Excel format.

        :return: an Excel file including the results of optimization problem.
        '''
        # Create a Workbook object
        workbook = Workbook()

        # Create a new sheet
        sheet = workbook.active

        # Get the columns to export excluding 'IsBestSolution', 'IsSolution', and 'pareto_pts'
        export_columns = [col for col in self.df.columns if col not in
                          ['IsBestSolution', 'IsSolution', 'pareto_pts']]

        # Write the column names to the sheet
        sheet.append(export_columns)

        # Write the row values to the sheet
        for _, row in self.df[export_columns].iterrows():
            values = row.tolist()
            sheet.append(values)

        # Apply conditional text color to the desired columns
        for column in self.specifications.index:
            current_spec = self.specifications.Type[column]
            if current_spec == "bounds":
                spec_values = self.specifications.Value[column]
                min_value, max_value = spec_values[0], spec_values[1]
                font_color = "000000"  # Default black color
                for index, value in self.df[column].items():
                    if value <= min_value:
                        font_color = "0000FF"  # Blue color
                    elif value >= max_value:
                        font_color = "FF0000"  # Red color
                    if column in export_columns:
                        cell = sheet.cell(row=index+2,
                                          column=export_columns.index(column)+1)
                        cell.font = Font(color=Color(rgb=font_color))
                        cell.value = value
            elif current_spec == "eq_cstr":
                font_color = "808080"  # Grey color
                for index in self.df.index:
                    if column in export_columns:
                        cell = sheet.cell(row=index+2,
                                          column=export_columns.index(column)+1)
                        cell.font = Font(color=Color(rgb=font_color))
                        cell.value = self.df[column][index]
            elif current_spec == "objective":
                font_color = "00FF00"  # Green color
                for index in self.df.index:
                    if column in export_columns:
                        cell = sheet.cell(row=index+2,
                                          column=export_columns.index(column)+1)
                        cell.font = Font(color=Color(rgb=font_color))
                        cell.value = self.df[column][index]

        # Save the workbook to a file
        workbook.save("exported_data.xlsx")
        print("Result saved!")

    def plot_one_selected_option(self,selected_option):
        '''
        Plots a selected variable in a graph.

        :param selected_option:
        :return:
        '''

        if isinstance(self.df[selected_option][0],(list,np.ndarray)): #vectorial
            index=[selected_option+str(i) for i in range(len(self.df
                                                [selected_option][0]))]
            df=pandas.DataFrame(np.array(self.df[selected_option].values.
                                         tolist()),columns=index)
            current_spec = self.specifications.Value[selected_option]
            current_type = self.specifications.Type[selected_option]

            if len(index)==2:
                self.fig,self.ax=plt.subplots(2,1)
            else:
                self.fig,self.ax=plt.subplots(2,2)
            self.ax=self.ax.ravel()

            for i in range(len(index)):
                element=index[i]
                if i!=0 and i%4==0:
                    self.display_figures(False)
                    self.fig, self.ax = plt.subplots(2, 2)
                    self.ax = self.ax.ravel()

                self.ax[i%4].plot(self.df['IterationNumber'],df[element],
                             label=element)
                self.ax[i%4].scatter(self.df['IterationNumber'], df[element])

            # Label each dot
                #for iteration, x, y in zip(self.df['IterationNumber'],
                #    self.df['IterationNumber'],df[element]):
                #    self.ax[i%4].annotate(iteration,(x,y),textcoords=
                #       "offset points",xytext=(0, 10), ha='center', va='bottom')

                current_spec_el=current_spec[index.index(element)]
                if current_type=="bounds":
                    self.ax[i%4].axhline(y=current_spec_el[1], color='red',
                                linestyle='--',label=element+'_max')
                    self.ax[i%4].axhline(y=current_spec_el[0],color='blue',
                                linestyle='--',label=element+'_min')

                elif current_type=='ineq_cstr':
                    if current_spec_el[1]!=None:
                        self.ax[i%4].axhline(y=current_spec_el[1],color='black',
                            linestyle='--',label=element + '_max')
                    if current_spec_el[0]!=None:
                        self.ax[i%4].axhline(y=current_spec_el[0], color='grey',
                            linestyle='--',label=element + '_min')

                elif current_type=='eq_cstr':
                    self.ax[i%4].axhline(y=current_spec_el, color='black',
                            linestyle='--',label=element+'_cstr')

                self.ax[i%4].grid()
                self.ax[i%4].set_xlabel("IterationNumber")
                self.ax[i%4].set_title(element)

                # Add the legend
                self.ax[i%4].legend()

                self.ax[i%4].xaxis.set_major_locator(MaxNLocator(integer=True))
                self.ax[i%4].ticklabel_format(useOffset=False, style='plain')

        else: #scalar
            self.fig, self.ax = plt.subplots()
            self.ax.plot(self.df['IterationNumber'],self.df[selected_option],
                    label=selected_option)
            self.ax.scatter(self.df['IterationNumber'],self.df[selected_option])

            # Label each dot
            #for iteration, x, y in zip(self.df['IterationNumber'],
            #            self.df['IterationNumber'], self.df[selected_option]):
            #    self.ax.annotate(iteration, (x, y), textcoords="offset points",
            #                xytext=(0, 10), ha='center', va='bottom')

            current_spec = self.specifications.Value[selected_option]
            current_type = self.specifications.Type[selected_option]

            if current_type=="bounds":
                self.ax.axhline(y=current_spec[1], color='red', linestyle='--',
                           label=selected_option+'_max')
                self.ax.axhline(y=current_spec[0], color='blue', linestyle='--',
                           label=selected_option+'_min')

            elif current_type=='ineq_cstr':
                if current_spec[1]!=None:
                    self.ax.axhline(y=current_spec[1], color='black',
                        linestyle='--',label=selected_option + '_max')
                if current_spec[0]!=None:
                    self.ax.axhline(y=current_spec[0], color='grey',
                        linestyle='--', label=selected_option + '_min')

            elif current_type=='eq_cstr':
                self.ax.axhline(y=current_spec, color='black', linestyle='--',
                           label=selected_option+'_cstr')

            self.ax.grid()
            self.ax.set_xlabel("IterationNumber")
            self.ax.set_title(selected_option)

            # Add the legend
            self.ax.legend()

            self.ax.xaxis.set_major_locator(MaxNLocator(integer=True))
            self.ax.ticklabel_format(useOffset=False, style='plain')

        self.display_figures(False)

    def display_figures(self,block=True):
        '''
        Displays figures of selected variables.

        :param block: bool indicating if figures must be closed before displaying others.
        :return: /
        '''
        #if self.canvas!=None:
        #    self.canvas.get_tk_widget().pack_forget()
        #    self.root.winfo_children()[-1].pack_forget()
        #self.canvas = FigureCanvasTkAgg(self.fig, master=self.root)  # A tk.DrawingArea.
        #self.canvas.get_tk_widget().pack(side=TOP, fill=BOTH, expand=1)
        #self.toolbar = NavigationToolbar2Tk(self.canvas, self.root)
        #self.canvas.draw_idle()
        plt.show(block=block)

    def plot_sorted_X(self):
        '''
        Plots output optimization variable w.r.t an input optimization variable graph.

        :return: /
        '''

        selected_x = self.selected_x.get()
        selected_y = self.selected_y.get()

        bbox=dict(boxstyle="round", fc="w", alpha=.3)
        arrowprops = dict(arrowstyle="->")

        if isinstance(self.df[selected_x][0],(list,np.ndarray))\
                and not isinstance(self.df[selected_y][0],(list,np.ndarray)):
            # vectorial_x and scalar_y
            index_x=[selected_x+str(i) for i in range(len(self.df
                                                [selected_x][0]))]
            x_values=pandas.DataFrame(np.array(self.df[selected_x].values.
                                         tolist()),columns=index_x)
            y_values=self.df[selected_y]

            if len(index_x) == 2:
                self.fig, self.ax = plt.subplots(2, 1)
            else:
                self.fig, self.ax = plt.subplots(2, 2)
            self.ax = self.ax.ravel()

            for i in range(len(index_x)):
                element_x=index_x[i]
                if i!=0 and i%4==0:
                    self.display_figures(False)
                    self.fig, self.ax = plt.subplots(2, 2)
                    self.ax = self.ax.ravel()
                self.ax[i%4].plot(x_values[element_x], y_values)
                self.ax[i%4].scatter(x_values[element_x], y_values)

                for iteration, x,y in zip(self.df['IterationNumber'],
                                  x_values[element_x],y_values):
                    self.ax[i%4].annotate(iteration,(x, y),textcoords=
                       "offset points",xytext=(0, 10), bbox=bbox,
                                          arrowprops=arrowprops)

                self.ax[i%4].grid()
                self.ax[i%4].set_xlabel(element_x)
                self.ax[i%4].set_ylabel(selected_y)
                self.ax[i%4].set_title(f"Sensitivity of {selected_y}/"
                                       f"{element_x}")

                # Avoid scientific notations
                self.ax[i%4].ticklabel_format(useOffset=False, style='plain')

        elif isinstance(self.df[selected_y][0],(list,np.ndarray))\
                and not isinstance(self.df[selected_x][0],(list,np.ndarray)):
            # vectorial_y and scalar_x
            index_y=[selected_y+str(i) for i in range(len(self.df
                                                [selected_y][0]))]
            y_values=pandas.DataFrame(np.array(self.df[selected_y].values.
                                         tolist()),columns=index_y)
            x_values=self.df[selected_x]

            if len(index_y)==2:
                self.fig,self.ax=plt.subplots(2,1)
            else:
                self.fig,self.ax=plt.subplots(2,2)
            self.ax=self.ax.ravel()

            for i in range(len(index_y)):
                element_y=index_y[i]
                if i!=0 and i%4==0:
                    self.display_figures(False)
                    self.fig, self.ax = plt.subplots(2, 2)
                    self.ax = self.ax.ravel()
                self.ax[i%4].plot(x_values, y_values[element_y])
                self.ax[i%4].scatter(x_values, y_values[element_y])

                for iteration, x,y in zip(self.df['IterationNumber'],
                                  x_values,y_values[element_y]):
                    self.ax[i%4].annotate(iteration,(x, y),textcoords=
                       "offset points",xytext=(0, 10), bbox=bbox,
                                          arrowprops=arrowprops)

                self.ax[i%4].grid()
                self.ax[i%4].set_xlabel(element_y)
                self.ax[i%4].set_ylabel(selected_y)
                self.ax[i%4].set_title(f"Sensitivity of {element_y}/"
                                       f"{selected_x}")

                # Avoid scientific notations
                self.ax[i%4].ticklabel_format(useOffset=False, style='plain')

        elif isinstance(self.df[selected_y][0],(list,np.ndarray))\
                and isinstance(self.df[selected_x][0],(list,np.ndarray)):
            # vectorial_y and vectorial_x
            index_x=[selected_x+str(i) for i in range(len(self.df
                                                [selected_x][0]))]
            index_y=[selected_y+str(i) for i in range(len(self.df
                                                [selected_y][0]))]
            y_values=pandas.DataFrame(np.array(self.df[selected_y].values.
                                         tolist()),columns=index_y)
            x_values=pandas.DataFrame(np.array(self.df[selected_x].values.
                                         tolist()),columns=index_x)

            self.fig, self.ax = plt.subplots(2,2)
            self.ax = self.ax.ravel()

            for i in range(len(index_x)):
                for j in range(len(index_y)):
                    n=i+j+i*(len(index_y)-1)
                    if (n!=0 and n%4==0):
                        self.display_figures(False)
                        self.fig, self.ax = plt.subplots(2, 2)
                        self.ax = self.ax.ravel()
                    element_x,element_y=index_x[i],index_y[j]
                    self.ax[n%4].plot(x_values[element_x], y_values[element_y])
                    self.ax[n%4].scatter(x_values[element_x],
                                         y_values[element_y])

                    for iteration, x,y in zip(self.df['IterationNumber'],
                                  x_values[element_x],y_values[element_y]):
                        self.ax[n%4].annotate(iteration,(x, y),textcoords=
                       "offset points",xytext=(0, 10), bbox=bbox,
                                          arrowprops=arrowprops)

                    self.ax[n%4].grid()
                    self.ax[n%4].set_xlabel(element_x)
                    self.ax[n%4].set_ylabel(element_y)
                    self.ax[n%4].set_title(
                        f"Sensitivity of {element_y}/{element_x}")

                    # Avoid scientific notations
                    self.ax[n%4].ticklabel_format(useOffset=False,style='plain')

        else: #scalar
            sorted_X_df = self.df.sort_values(by=[selected_x])

            # Get the arrays of values for the selected options
            x_values = sorted_X_df[selected_x].values
            y_values = sorted_X_df[selected_y].values

            self.fig, self.ax = plt.subplots()
            self.ax.plot(x_values, y_values)
            self.ax.scatter(x_values, y_values)

            for iteration, x,y in zip(sorted_X_df['IterationNumber'],
                                  x_values,y_values):
                self.ax.annotate(iteration, (x, y), textcoords="offset points",
                        xytext=(0, 10), bbox=bbox,arrowprops=arrowprops)

            self.ax.grid()
            self.ax.set_xlabel(selected_x)
            self.ax.set_ylabel(selected_y)
            self.ax.set_title(f"Sensitivity of {selected_y} / {selected_x}")

            # Avoid scientific notations
            self.ax.ticklabel_format(useOffset=False, style='plain')

        self.display_figures()

    def exit_program(self):
        '''
        Stops the Graphical User Interface.

        :return: /
        '''
        sys.exit()


# SPDX-FileCopyrightText: 2021 G2Elab / MAGE
#
# SPDX-License-Identifier: Apache-2.0

from scipy.stats import qmc,norm
from jax import config
from noloadj.analyse.DGSM import computeParametricValue,computeParametricJacobian
from typing import Callable, Dict, List, Any
from pandas import DataFrame as df
import numpy as np,sys
config.update("jax_enable_x64", True)

'''1st_order Sobol indices'''

def computeSobol(model: Callable[..., Dict], inputs: Dict[str, Any],
                outputs: List[str], deltas:Dict[str,Any],
                deltas_type:Dict[str,Any],N:int,Param: Dict[str, Any]={}):
    '''
    Compute 1st-order Sobol indices of a model.

    :param model: model
    :param inputs: dict: names and initial values of derivable input variables
    :param outputs: list: names of output variables
    :param deltas: dict: variation of derivable input variables
    :param deltas_type: dict: type of variation for each derivable input ('rel' for relative or 'abs' for absolute)
    :param N: int: number of points to compute DGSM. Must be a power of 2.
    :param Param: names and values of non-derivable input variables (dictionary)
    :return: dataframe : Sobol indices
    '''
    if int(bin(N)[3:])!=0: # test si N est une puissance de 2
        print('Warning :', N, 'is not a power of 2 (512, 1024, 2058, ...)')
        sys.exit(0)

    inputs_names,init_values=list(inputs.keys()),list(inputs.values())
    len_inputs,len_outputs=len(inputs_names),len(outputs)
    step=2*len_inputs+2
    nb=int(step*N)

    Si=np.empty((len_inputs,len_outputs))

    sampled=sample(len_inputs,N) #= qmc.Sobol(d=len_inputs, scramble=False).random(nb) # création des échantillons [0,1]
    l_bounds,u_bounds=[],[]

    for input in inputs_names:
        index_input=inputs_names.index(input)
        if deltas_type[input]=='abs':
            l_bounds.append(init_values[index_input]-deltas[input])
            u_bounds.append(init_values[index_input]+deltas[input])
        elif deltas_type[input]=='rel':
            l_bounds.append(init_values[index_input]-deltas[input]*
                            init_values[index_input])
            u_bounds.append(init_values[index_input]+deltas[input]*
                            init_values[index_input])

    scale_samples=qmc.scale(sampled, l_bounds,u_bounds).T # mise à échelle des échantillons

    for input in inputs_names:
        index_input=inputs_names.index(input)
        inputs[input] = scale_samples[index_input]
                                  
    outputs_values = computeParametricValue(model,inputs,outputs,nb,Param) # évaluation modèle nb fois

    for j in range(len_outputs):
        out=outputs_values[j]
        out=(out-np.mean(out))/np.std(out)
        A=out[0:nb:step]
        B=out[step-1:nb:step]
        var=np.var(np.r_[A, B])
        for i in range(len_inputs):
            AB=out[i+1:nb:step]
            res=B*(AB-A)
            Si[i,j]=np.mean(res)/var

    Sobol=df(Si,index=inputs_names,columns=outputs)

    return Sobol


def sample(len_inputs,N):
    base_sequence = qmc.Sobol(d=2*len_inputs).random(N)
    sobol_sequence = np.zeros([(2*len_inputs + 2) * N, len_inputs])
    index = 0

    for i in range(N):
        for j in range(len_inputs):
            sobol_sequence[index,j] = base_sequence[i,j]
        index += 1

        for k in range(len_inputs):
            for j in range(len_inputs):
                if j == k:
                    sobol_sequence[index,j] = base_sequence[i,j+len_inputs]
                else:
                    sobol_sequence[index,j] = base_sequence[i,j]
            index += 1

        for k in range(len_inputs):
            for j in range(len_inputs):
                if j == k:
                    sobol_sequence[index,j] = base_sequence[i,j]
                else:
                    sobol_sequence[index,j] = base_sequence[i,j+len_inputs]

            index += 1

        for j in range(len_inputs):
            sobol_sequence[index,j] = base_sequence[i,j+len_inputs]
        index += 1

    return sobol_sequence



# SPDX-FileCopyrightText: 2021 G2Elab / MAGE
#
# SPDX-License-Identifier: Apache-2.0

from noloadj.optimization.iterationHandler import Iterations
from typing import Callable, Dict, List, Any
import numpy,sys
from jax import config
config.update("jax_enable_x64", True)
import jax.numpy as jnp

'''computes model with the inputs dictionnary, and give the value of output 
variables'''
def computeOnce(model: Callable[..., Dict], inputs: Dict[str, Any],
                outputs: List[str]):
    """
    Computes the outputs of the model according to the values of the given
    input variables.

    :param model: model
    :param inputs: names and values of input variables (dictionary)
    :param outputs: names of output variables (list)
    :return: values of the outputs (list)
    """
    res = model(**inputs)
    dico = {k: v for k, v in res.__iter__()}  # conversion en dictionnaire
    out=[]
    nans_vars=[]
    for vars in outputs:
        try:
            if vars not in list(dico.keys()):
                raise KeyError(vars) #si la variable du cahier des charges
            if jnp.all(jnp.isnan(dico[vars])):
                raise ValueError(vars)
        except KeyError: # n'appartient pas aux sorties du modèle
            print('Warning :',vars,'is not in model')
            sys.exit(0)
        except ValueError:
            nans_vars.append(vars)
        else:
            out.append(dico[vars])
    # sauvegarde et tracé des résultats uniquement si appel à la fonction et
    # non au gradient
    if nans_vars!=[]:
        print('Warning :',nans_vars,'is Nan')
        sys.exit(0)
    for i in range(len(out)): # POUR RETIRER LES DEVICE ARRAY
        if numpy.size(out[i])==1:
            out[i]=float(out[i])
        #else:
        #    for j in range(len(out[i])):
        #        out[i][j]=float(out[i][j])
    return out

#TODO : Harmoniser les résultats avec ceux d'une optimisation pour le tracé
# des itérations
'''computes model with the inputs dictionnary and a variable input varying in 
a range, and give the value of output variables'''
def computeParametric(model: Callable[..., Dict], variable: str,
                      range: List[float], inputs: Dict[str, Any],
                      outputs: List[str]):
    """
    Computes the outputs of the model corresponding to the values of the given
    input variables, except one varying in a range of values.

    :param model: model
    :param variable: the input that varies
    :param range: range of values in which the input varies
    :param inputs: names and values of constant input variables (dictionary)
    :param outputs: names of output variables (list)
    :return: values of the outputs (list)
    """
    iter = Iterations(None,[variable], outputs)  # permet de sauvegarder les
    # résultats au fur et à mesure (optionnel)
    for x in range:
        if inputs!=[]:
            res = model(**{variable: x}, **inputs)
        else:
            res = model(*x) # in case of model with only 1 argument
        dico = {k: v for k, v in res.__iter__()}  # conversion en dictionnaire
        out=[]
        nans_vars = []
        for vars in outputs:
            try:
                if vars not in list(dico.keys()):
                    raise KeyError(vars) #si la variable du cahier des charges
                if jnp.all(jnp.isnan(dico[vars])):
                    raise ValueError(vars)
            except KeyError: # n'appartient pas aux sorties du modèle
                print('Warning :',vars,'is not in model')
                sys.exit(0)
            except ValueError:
                nans_vars.append(vars)
            else:
                out.append(dico[vars])
        # sauvegarde et tracé des résultats uniquement si appel à la fonction et
        # non au gradient
        if nans_vars != []:
            print('Warning :', nans_vars, 'is Nan')
            sys.exit(0)
        iter.updateData([x], out)
    return iter


from jax import jacfwd

def computeJacobian(model: Callable[..., Dict], inputs: Dict[str, Any],
                      outputs: List[str],Param: Dict[str, Any]={}):
    """
    Displays gradients of outputs corresponding to the values of
    the given inputs.

    :param model: model
    :param inputs: names and values of derivable input variables (dictionary)
    :param outputs: names of output variables (list)
    :param Param: names and values of non-derivable input variables (dictionary)
    :return: gradients of the outputs (list)
    """
    inputs_keys=list(inputs.keys())
    inputs_values=list(inputs.values())
    def eval_model(inputs_values):
        xList=dict(zip(inputs_keys,inputs_values))
        if Param!={}:
            res = model(**xList,**Param)
        else:
            res = model(**xList)
        dico = {k: v for k, v in res.__iter__()}  # conversion en dictionnaire
        out=[]
        nans_vars = []
        for vars in outputs:
            try:
                if vars not in list(dico.keys()):
                    raise KeyError(vars) #si la variable du cahier des charges
                if jnp.all(jnp.isnan(dico[vars])):
                    raise ValueError(vars)
            except KeyError: # n'appartient pas aux sorties du modèle
                print('Warning :',vars,'is not in model')
                sys.exit(0)
            except ValueError:
                nans_vars.append(vars)
            else:
                out.append(dico[vars])
        if nans_vars != []:
            print('Warning :', nans_vars, 'is Nan')
            sys.exit(0)
        return out

    nans_vars = []
    dout=jacfwd(eval_model)(inputs_values)
    for i in range(len(dout)): # POUR RETIRER LES DEVICE ARRAY
        for j in range(len(dout[i])):
            if numpy.size(dout[i][j])==1:
                dout[i][j]=float(dout[i][j])
            else:
                for k in range(len(dout[i][j])):
                    dout[i][j][k]=float(dout[i][j][k])
        try:
            if numpy.all(numpy.isnan(dout[i])):
                raise ValueError(i)
        except ValueError:
            nans_vars.append(outputs[i])
    if nans_vars!=[]:
        print('Warning :', nans_vars, 'is Nan')
        print('in Gradient Computation')
        sys.exit(0)
    return dout


from xml.dom import minidom
from noloadj.optimization.wrapper import Iterations

def resultsToXML(iter: Iterations, fileName):
    """
    Return an XMLfile compatible with CADES. This can be used to plot geometry
    in GeomMaker.

    :param fileName: the filename to save XML tree
    :return: /
    """
    iterNb = len(iter.solutions)
    root = minidom.Document()
    xml = root.createElement('Optimization')
    root.appendChild(xml)
    #<Iterations Number="12">
    iterations = root.createElement('Iterations')
    iterations.setAttribute('Number', str(iterNb))
    xml.appendChild(iterations)

    #<Iteration Number="0" isBestSolution="false" isSolution="false">
    number = 0  # numéro de l'itération
    for sol in iter.solutions:
        if len(iter.iNames) == iterNb:
            isBestSolution="true"   #on suppose qu'un algo type SQP est utilisé
            # conduisant à avoir la meilleure solution à la dernière itération
        else:
            isBestSolution="false"
        isSolution = "true" #true par défaut, il faudrait verifier si les
        # contraintes sont violées ou pas.
        iteration = root.createElement('Iteration')
        iteration.setAttribute('Number', str(number))
        iteration.setAttribute('isBestSolution', isBestSolution)
        iteration.setAttribute('isSolution', isSolution)
        iterations.appendChild(iteration)
        number = number + 1

        #<Inputs Number="22">
        inputs = root.createElement('Inputs')
        inputs.setAttribute('Number', str(len(iter.iNames)))
        iteration.appendChild(inputs)
        iCmpt = 0
        #<Input Value="250000.0" Name="S"/>
        for inputVal in sol.iData:
            input = root.createElement('Input')
            input.setAttribute('Value', str(inputVal))
            input.setAttribute('Name', iter.iNames[iCmpt])
            inputs.appendChild(input)
            iCmpt=iCmpt+1

        #<Outputs Number="116">
        outputs = root.createElement('Outputs')
        outputs.setAttribute('Number', str(len(iter.oNames)+len(iter.fNames)))
        iteration.appendChild(outputs)
        oCmpt = 0
        #<Output Value="1.25663704E-6" Name="mu0"/>
        for outputVal in sol.oData:
            output = root.createElement('Output')
            output.setAttribute('Value', str(outputVal))
            output.setAttribute('Name', iter.oNames[oCmpt])
            outputs.appendChild(output)
            oCmpt=oCmpt+1
        oCmpt = 0
        for outputVal in sol.fData:
            output = root.createElement('Output')
            output.setAttribute('Value', str(outputVal))
            output.setAttribute('Name', iter.fNames[oCmpt])
            outputs.appendChild(output)
            oCmpt=oCmpt+1
    #< SPECIFICATIONS >
    spec = root.createElement('SPECIFICATIONS') #not used but required
    xml.appendChild(spec)

    bnds = root.createElement('BoundsOfInputs')
    bnds.setAttribute('Number',str(len(iter.bounds)))
    spec.appendChild(bnds)

    for name,value in iter.bounds.items():
        bnd = root.createElement('Bounds')
        bnd.setAttribute('Type', 'Input')
        bnd.setAttribute('Value', str(value))
        bnd.setAttribute('Name', name)
        bnds.appendChild(bnd)

    objs = root.createElement('ObjectiveFunctions')
    objs.setAttribute('Number',str(len(iter.objectives)))
    spec.appendChild(objs)

    for name,value in iter.objectives.items():
        obj = root.createElement('Objective')
        obj.setAttribute('Type', 'Output')
        obj.setAttribute('Value', str(value))
        obj.setAttribute('Name', name)
        objs.appendChild(obj)

    if iter.eq_cstr!={}:
        eqs = root.createElement('EqualityConstraints')
        eqs.setAttribute('Number', str(len(iter.eq_cstr)))
        spec.appendChild(eqs)

        for name,value in iter.eq_cstr.items():
            eq = root.createElement('EqualityConstraint')
            eq.setAttribute('Type', 'Output')
            eq.setAttribute('Value', str(value))
            eq.setAttribute('Name', name)
            eqs.appendChild(eq)

    if iter.ineq_cstr!={}:
        ineqs = root.createElement('InequalityConstraints')
        ineqs.setAttribute('Number', str(len(iter.ineq_cstr)))
        spec.appendChild(ineqs)

        for name, value in iter.ineq_cstr.items():
            ineq = root.createElement('InequalityConstraint')
            ineq.setAttribute('Type', 'Output')
            ineq.setAttribute('Value', str(value))
            ineq.setAttribute('Name', name)
            ineqs.appendChild(ineq)

    if iter.fNames!=[]:
        fOuts = root.createElement('FreeOutputs')
        fOuts.setAttribute('Number', str(len(iter.fNames)))
        spec.appendChild(fOuts)

        for name in iter.fNames:
            fOut=root.createElement('FreeOutput')
            fOut.setAttribute('Type', 'Output')
            fOut.setAttribute('Name', name)
            fOuts.appendChild(fOut)

    xml_str = root.toprettyxml(indent ="\t")
    with open(fileName, "w") as f:
        f.write(xml_str)



import jax.numpy as np
from jax import jvp,lax
from abc import abstractmethod
#from noloadj.numerical_methods.ODE.ode_tools import Switch

################################################################################
# TIME FEATURES
def T_pair(T):
    return lambda t:(t//T)%2==0

def T_impair(T):
    return lambda t:(t//T)%2!=0

def T_numero(T,n,i):
    return lambda t:(t//T)%n!=i

class TimeFeature:
    '''
    Interface class to implement time-domain criteria to extract from dynamic simulation.

    Attributes:
        - name : str to explicit on which variable the criteria will be applied
        - type : str to explicit its type (time-domain criteria 'time')
    '''
    def __init__(self,name):
        self.name=name
        self.type='time'
    @abstractmethod
    def init(self,z0,t0,h0):
        '''
        Compute the initial value of the criteria.

        :param z0: float : the initial value of the variable given in "name" attribute
        :param t0: float : time for the initial iteration
        :param h0: float : step size for the initial iteration
        :return: The initial value of the criteria
        '''
        pass
    @abstractmethod
    def expression(self,t_prev,z_prev,t_now,z_now,cstr,h_prev,T):
        '''
        Compute the value of the criteria for the present iteration of the simulation.

        :param t_prev: float : time for the previous iteration
        :param z_prev: the value of the variable given in "name" attribute, for the previous iteration
        :param t_now: float : time for the  present iteration
        :param z_now: float : the value of the variable given in "name" attribute, for the present iteration
        :param cstr: float : the criteria value for the  previous iteration
        :param h_prev: float : step size for the  previous iteration
        :param T: float : operating period
        :return: The value of the criteria for the present iteration of the simulation
        '''
        pass
    @abstractmethod
    def fin(self,tf,cstr,T):
        '''
        Compute the final value of the criteria.

        :param tf: float : final time of simulation
        :param cstr: float : the criteria value before for the previous iteration of final time
        :param T: float : operating period
        :return: The final value of the criteria
        '''
        pass
    @abstractmethod
    def dinit(self,z0,dz0,t0,h0):
        '''
        Compute the initial gradients of the criteria.

        :param z0: float : the initial value of the variable given in "name" attribute
        :param dz0: float : the initial gradients of the variable given in "name" attribute
        :param t0: float : time for the initial iteration
        :param h0: float : step size for the initial iteration
        :return: The initial gradients of the criteria
        '''
        pass
    @abstractmethod
    def der_expression(self,t_prev,z_prev,dz_prev,t_now,z_now,dz_now,cstr,dcstr,
                    h_prev,T,dT):
        '''
        Compute the gradients of the criteria for the present iteration of the simulation.

        :param t_prev: float : time for the previous iteration
        :param z_prev: float : the value of the variable given in "name" attribute, for the previous iteration
        :param dz_prev: Array : the gradients of the variable given in "name" attribute, for previous iteration
        :param t_now: float : time for the present iteration
        :param z_now: float : the value of the variable given in "name" attribute, for present iteration
        :param dz_now: Array : the gradients of the variable given in "name" attribute, for present iteration
        :param cstr: float : the criteria value for the previous iteration
        :param dcstr: Array : the criteria gradients for the previous iteration
        :param h_prev: float : step size for the previous iteration
        :param T: float : operating period
        :param dT: Array : gradients of operating period
        :return: The gradients of the criteria for the present iteration of the simulation
        '''
        pass
    @abstractmethod
    def der_fin(self,tf,cstr,T,dcstr,dT,zf):
        '''
        Compute the final gradients of the criteria.

        :param tf: float : final time of simulation
        :param cstr: float : the criteria value before for the previous iteration of final time
        :param T: float : operating period
        :param dcstr: Array : the criteria gradients before for the previous iteration of final time
        :param dT: Array : gradients of operating period
        :param zf: float : final value of the variable given in "name" attribute
        :return: The final gradients of the criteria
        '''
        pass


class Min(TimeFeature):
    '''
    Minimum of a variable over a period of dynamic simulation.
    '''
    def __init__(self,name,nbT=1):
        TimeFeature.__init__(self,name)
        self.nbT=nbT
    def init(self,z0,t0,h0):
        return z0
    def expression(self,t_prev,z_prev,t_now,z_now,cstr,h_prev,T):
        ind=np.where(T==0.,0,1)
        def without_T():
            return np.minimum(z_now,cstr)
        def with_T():
            return np.where((t_prev//(self.nbT*T))==(t_now//(self.nbT*T)),
                        np.minimum(z_now,cstr),z_now)
        return lax.switch(ind,[without_T,with_T])
    def fin(self,tf,cstr,T):
        return cstr
    def dinit(self,z0,dz0,t0,h0):
        return dz0
    def der_expression(self,t_prev,z_prev,dz_prev,t_now,z_now,dz_now,cstr,dcstr,
                    h_prev,T,dT):
        return jvp(self.expression,(t_prev,z_prev,t_now,z_now,cstr,h_prev,T),
                   (0.,dz_prev,0.,dz_now,dcstr,0.,dT))[1]
    def der_fin(self,tf,cstr,T,dcstr,dT,zf):
        return dcstr

class Max(TimeFeature):
    '''
    Maximum of a variable over a period of dynamic simulation.
    '''
    def __init__(self,name,nbT=1):
        TimeFeature.__init__(self,name)
        self.nbT=nbT
    def init(self,z0,t0,h0):
        return z0
    def expression(self,t_prev,z_prev,t_now,z_now,cstr,h_prev,T):
        ind=np.where(T==0.,0,1)
        def without_T():
            return np.maximum(z_now,cstr)
        def with_T():
            return np.where((t_prev//(self.nbT*T))==(t_now//(self.nbT*T)),
                        np.maximum(z_now,cstr),z_now)
        return lax.switch(ind,[without_T,with_T])
    def fin(self,tf,cstr,T):
        return cstr
    def dinit(self,z0,dz0,t0,h0):
        return dz0
    def der_expression(self,t_prev,z_prev,dz_prev,t_now,z_now,dz_now,cstr,dcstr,
                    h_prev,T,dT):
        return jvp(self.expression,(t_prev,z_prev,t_now,z_now,cstr,h_prev,T),
                   (0.,dz_prev,0.,dz_now,dcstr,0.,dT))[1]
    def der_fin(self,tf,cstr,T,dcstr,dT,zf):
        return dcstr

class moy(TimeFeature):
    '''
    Average value of a variable over a period of dynamic simulation.
    '''
    def __init__(self,name):
        TimeFeature.__init__(self, name)
    def init(self,z0,t0,h0):
        return 0.
    def expression(self,t_prev,z_prev,t_now,z_now,cstr,h_prev,T):
        ind=np.where(T==0.,0,1)
        def without_T():
            return cstr+0.5*h_prev*(z_prev+z_now)
        def with_T():
            return np.where((t_prev//T)==(t_now//T),cstr+0.5*h_prev*(z_prev+
                                                                z_now),0.)
        return lax.switch(ind,[without_T,with_T])
    def fin(self,tf,cstr,T):
        ind=np.where(T==0.,0,1)
        def without_T():
            return cstr/tf
        def with_T():
            return cstr/T
        return lax.switch(ind,[without_T,with_T])
    def dinit(self,z0,dz0,t0,h0):
        return 0.
    def der_expression(self,t_prev,z_prev,dz_prev,t_now,z_now,dz_now,cstr,dcstr,
                    h_prev,T,dT):
        ind=np.where(T==0.,0,1)
        def without_T():
            return dcstr+0.5*h_prev*(dz_prev+ dz_now)
        def with_T():
            return np.where((t_prev//T)==(t_now//T),dcstr+0.5*h_prev*(dz_prev+
                                                                  dz_now),0.)
        return lax.switch(ind,[without_T,with_T])
    def der_fin(self,tf,cstr,T,dcstr,dT,zf):
        ind=np.where(T==0.,0,1)
        def without_T():
            return dcstr/tf
        def with_T():
            return dcstr/T+((zf-cstr)/T)*dT
        return lax.switch(ind,[without_T,with_T])

class eff(TimeFeature):
    '''
    RMS value of a variable over a period of dynamic simulation.
    '''
    def __init__(self,name):
        TimeFeature.__init__(self,name)
    def init(self,z0,t0,h0):
        return 0.
    def expression(self,t_prev,z_prev,t_now,z_now,cstr,h_prev,T):
        ind=np.where(T==0.,0,1)
        def without_T():
            return cstr+0.5*h_prev*(z_prev**2+ z_now**2)
        def with_T():
            return np.where((t_prev//T)==(t_now//T),cstr+0.5*h_prev*(z_prev**2+
                                                                 z_now**2),0.)
        return lax.switch(ind,[without_T,with_T])
    def fin(self,tf,cstr,T):
        ind=np.where(T==0.,0,1)
        def without_T():
            return np.sqrt(cstr/tf)
        def with_T():
            return np.sqrt(cstr/T)
        return lax.switch(ind,[without_T,with_T])
    def dinit(self,z0,dz0,t0,h0):
        return 0.
    def der_expression(self,t_prev,z_prev,dz_prev,t_now,z_now,dz_now,cstr,dcstr,
                    h_prev,T,dT):
        ind=np.where(T==0.,0,1)
        def without_T():
            return dcstr+0.5*h_prev*(2*z_prev*dz_prev+2*z_now*dz_now)
        def with_T():
            return np.where((t_prev//T)==(t_now//T),dcstr+0.5*h_prev*(2*z_prev*
                                    dz_prev+2*z_now*dz_now),0.)
        return lax.switch(ind,[without_T,with_T])
    def der_fin(self,tf,cstr,T,dcstr,dT,zf):
        ind=np.where(T==0.,0,1)
        def without_T():
            return dcstr/(2*tf*cstr)
        def with_T():
            return dcstr/(2*T*cstr)+(zf**2-cstr**2)/(2*cstr*T)*dT
        return lax.switch(ind,[without_T,with_T])

################################################################################
# FREQUENCY FEATURES

class FreqFeature:
    '''
    Interface class to implement frequency-domain criteria to extract from dynamic simulation.

    Attributes:
        - name : str to explicit on which variable the criteria will be applied.
        - type : str to explicit its type (frequency-domain criteria 'freq')
    '''
    def __init__(self,name):
        self.name=name
        self.type='freq'
    @abstractmethod
    def expression(self,module,phase,vect_freq,f):
        """
        Compute the value of the criteria.

        :param module: Array : the module of the FFT of the variable given in "name" attribute
        :param phase: Array : the phase of the FFT of the variable given in "name" attribute
        :param vect_freq: Array : frequency vector
        :param f: float : fundamental frequency
        :return: Value of the criteria
        """
        pass
    @abstractmethod
    def der_expression(self,module,phase,dmodule,dphase,vect_freq,f):
        """
        Compute the gradients of the criteria with respect to optimization inputs.

        :param module: Array : the module of the FFT of the variable given in "name" attribute
        :param phase:  Array : the phase of the FFT of the variable given in "name" attribute
        :param dmodule: Array : the gradients of the module of the FFT of the variable given in "name" attribute
        :param dphase: Array : the gradients of the phase of the FFT of the variable given in "name" attribute
        :param vect_freq: Array : frequency vector
        :param f: float : fundamental frequency
        :return: Gradients of the criteria
        """
        pass



class Module_FFT(FreqFeature):
    '''
    Module of the FFT of a variable for a given frequency.
    '''
    def __init__(self,name,number):
        FreqFeature.__init__(self,name)
        self.number=number
    def expression(self,module,phase,vect_freq,f):
        if isinstance(self.number,int):
            if self.number==0:
                return module[0]
            elif self.number==1:
                indf=np.argmin(np.abs(vect_freq-f))
                return module[indf]
            else:
                res=np.zeros(self.number)
                for j in range(len(res)):
                    indf=np.argmin(np.abs(vect_freq-j*f))
                    res=res.at[j].set(module[indf])
        else:
            res=np.zeros(len(self.number))
            for j in range(len(res)):
                indf=np.argmin(np.abs(vect_freq-self.number[j]))
                res=res.at[j].set(module[indf])
        return res
    def der_expression(self,module,phase,dmodule,dphase,vect_freq,f):
        if isinstance(self.number,int):
            if self.number==0:
                return module[0],dmodule[0]
            elif self.number==1:
                indf=np.argmin(np.abs(vect_freq-f))
                return module[indf],dmodule[1]
            else:
                res = np.zeros(self.number)
                dres=np.zeros(self.number)
                for j in range(len(res)):
                    indf=np.argmin(np.abs(vect_freq-j*f))
                    res=res.at[j].set(module[indf])
                    dres=dres.at[j].set(dmodule[indf])
        else:
            res = np.zeros(len(self.number))
            dres=np.zeros(len(self.number))
            for j in range(len(res)):
                indf=np.argmin(np.abs(vect_freq-self.number[j]))
                res=res.at[j].set(module[indf])
                dres=dres.at[j].set(dmodule[indf])
        return res,dres

class THD(FreqFeature):
    '''
    Total Harmonic Distorsion (THD) of a variable.
    '''
    def __init__(self,name):
        FreqFeature.__init__(self, name)
    def expression(self,module,phase,vect_freq,f):
        ref=np.maximum(module[0],module[1])
        harm=module[2::]
        THD=np.sqrt(np.sum(harm**2))/ref
        return THD
    def der_expression(self,module,phase,dmodule,dphase,vect_freq,f):
        ref=np.maximum(module[0],module[1])
        dref=np.where(ref==module[0],dmodule[0],dmodule[1])
        harm,dharm=module[2::],dmodule[2::]
        THD=np.sqrt(np.sum(harm**2))/ref
        dTHD=-dref*THD/ref+np.sum(np.dot(harm,dharm))/(ref*ref*THD)
        return THD,dTHD




import tkinter as tk
from tkinter import ttk
from noloadj.gui.gui1 import App as App1
import xml.etree.ElementTree as ET
import pandas as pd
from ast import literal_eval


def preprocess_xml(xml_file_path):
    '''
    Returns information of the XML optimization results file to the GUI.

    :param xml_file_path: str: name of the xml file
    :return: Several outputs:
    - df : dataframe: results of the optimization problem towards iterations.
    - option_list : list: names of optimization inputs and outputs.
    - specifications_df : dataframe for specifications
    - objectives: dataframe for values of the objective functions towards optimization.
    '''
    # Parse the XML file
    tree = ET.parse(xml_file_path)
    root = tree.getroot()

    # Extract data from XML
    data = []
    for iteration in root.iter("Iteration"):
        iteration_number = int(iteration.get("Number"))
        is_best_solution = iteration.get("isBestSolution")
        is_solution = iteration.get("isSolution")
        inputs = iteration.find("Inputs")
        outputs = iteration.find("Outputs")

        inputs_data = {input_elem.get("Name"): float(input_elem.get("Value"))
                       for input_elem in inputs.iter("Input")}
        outputs_data = {output_elem.get("Name"): float(output_elem.get("Value"))
                        for output_elem in outputs.iter("Output")}

        iteration_data = {
            "IterationNumber": iteration_number,
            "IsBestSolution": is_best_solution,
            "IsSolution": is_solution,
            **inputs_data,
            **outputs_data
        }
        data.append(iteration_data)

    # Convert data to DataFrame
    df = pd.DataFrame(data)
    option_list = df.columns[3:].values.tolist()

    # Extract specifications from XML
    bounds = root.find("SPECIFICATIONS/BoundsOfInputs")
    objective_functions = root.find("SPECIFICATIONS/ObjectiveFunctions")
    equality_constraints = root.find("SPECIFICATIONS/EqualityConstraints")
    inequality_constraints = root.find("SPECIFICATIONS/InequalityConstraints")
    free_outputs = root.find("SPECIFICATIONS/FreeOutputs")

    all_dfs_array = []

    # Extract bounds
    bounds_data = []
    for bound in bounds.iter("Bounds"):
        bound_name = bound.get("Name")
        bound_value = bound.get("Value")
        bound_in_out = bound.get("Type")
        bounds_data.append(
            {"Name": bound_name, "Value": literal_eval(bound_value),
             "In_Out": bound_in_out})

    df_bounds = pd.DataFrame(bounds_data)
    df_bounds["Type"] = "bounds"

    all_dfs_array.append(df_bounds)

    # Extract objective function data
    objective_functions_data = []
    for obj_func in objective_functions.iter("Objective"):
        objective_name = obj_func.get("Name")
        objective_value = obj_func.get("Value")
        objective_in_out = obj_func.get("Type")
        objective_functions_data.append({"ObjectiveName": objective_name,
                                         "ObjectiveValue": literal_eval(
                                             objective_value),
                                         "In_Out": objective_in_out})

    # Convert objective function data to DataFrame
    df_objective_functions = pd.DataFrame(objective_functions_data)
    df_objective_functions["Type"] = "objective"
    df_objective_functions.columns = ["Name", "Value", "In_Out", "Type"]

    objectives = df_objective_functions["Name"].values

    all_dfs_array.append(df_objective_functions)

    # Extract equality constraint data
    if equality_constraints is not None:
        equality_constraints_data = []
        for eq_constraint in equality_constraints.iter("EqualityConstraint"):
            constraint_name = eq_constraint.get("Name")
            constraint_value = eq_constraint.get("Value")
            constraint_in_out = eq_constraint.get("Type")
            equality_constraints_data.append({"ConstraintName": constraint_name,
                                              "ConstraintValue": literal_eval(
                                                  constraint_value),
                                              "In_Out": constraint_in_out})

        # Convert equality constraint data to DataFrame
        df_equality_constraints = pd.DataFrame(equality_constraints_data)

        df_equality_constraints["Type"] = "eq_cstr"
        df_equality_constraints.columns = ["Name", "Value", "In_Out", "Type"]

        all_dfs_array.append(df_equality_constraints)

    # Extract inequality constraint data
    if inequality_constraints is not None:
        inequality_constraints_data = []
        for ineq_constraint in inequality_constraints.iter(
                "InequalityConstraint"):
            constraint_name = ineq_constraint.get("Name")
            constraint_value = ineq_constraint.get("Value")
            constraint_in_out = ineq_constraint.get("Type")
            inequality_constraints_data.append(
                {"ConstraintName": constraint_name,
                 "ConstraintValue": literal_eval(constraint_value),
                 "In_Out": constraint_in_out})

        # Convert inequality constraint data to DataFrame
        df_inequality_constraints = pd.DataFrame(inequality_constraints_data)

        df_inequality_constraints["Type"] = "ineq_cstr"
        df_inequality_constraints.columns = ["Name", "Value", "In_Out", "Type"]

        all_dfs_array.append(df_inequality_constraints)

    # Extract free output data
    if free_outputs is not None:
        free_outputs_data = []
        for free_output in free_outputs.iter("FreeOutput"):
            output_name = free_output.get("Name")
            free_in_out = free_output.get("Type")
            free_outputs_data.append(
                {"OutputName": output_name, "In_Out": free_in_out})

        # Convert free output data to DataFrame
        df_free_outputs = pd.DataFrame(free_outputs_data)

        df_free_outputs["Type"] = "free"
        df_free_outputs.columns = ["Name", "In_Out", "Type"]

        # Filter out the objectives
        df_free_outputs = df_free_outputs[
            ~df_free_outputs["Name"].isin(objectives)]

        all_dfs_array.append(df_free_outputs)

    specifications_df = pd.DataFrame(
        columns=["Name", "Value", "In_Out", "Type"])

    for i in all_dfs_array:
        specifications_df = pd.concat([specifications_df, i], axis=0)

    specifications_df = specifications_df.set_index('Name')

    return df, option_list, specifications_df, objectives


def openGUI(result):
    """
    Opens Graphical User Interface for post-processing optimization.

    :param result: str or Iterations class : a XML file or an instance of Iterations class that contains values of optimization variables across iterations.
    :return: /
    """
    if isinstance(result,str):
        df,option_list,specifications_df,objectives=preprocess_xml(result)
    else:
        df=result.print()
        df["IsBestSolution"]='false'
        df["IsSolution"] = 'true'
        df["IterationNumber"]=[i for i in range(len(df))]
        option_list=result.iNames+result.oNames+result.fNames
        specifications_df=result.printSpec()

    ### Tk App ###

    root = tk.Tk()
    root.title("Notebook")

    # Create a tabbed interface
    notebook = ttk.Notebook(root)
    notebook.pack(fill=tk.BOTH, expand=True)

    # Create the first tab and add the App and ImageWindow instances from program1
    tab1 = ttk.Frame(notebook)
    app1 = App1(tab1, option_list, df, specifications_df)
    notebook.add(tab1)


    root.mainloop()


# SPDX-FileCopyrightText: 2021 G2Elab / MAGE
#
# SPDX-License-Identifier: Apache-2.0

from typing import Dict, Callable
import scipy.optimize as scipyOpt
from noloadj.optimization.specifications import Spec
from noloadj.optimization.wrapper import Wrapper
from noloadj.optimization.iterationHandler import Iterations
from noloadj.optimization.multiobjective import EpsilonConstraint
from noloadj.optimization.Tools import *
import numpy
import jax.numpy as jnp

class OptimProblem:
    # TODO : standardiser les results d'optim entre 1D : SLSQP Result et
    #  2D : Pareto...
    """
    Class used to define the optimization problem.

    Attributes :

    - model : Callable: function which represents the optimization loop.
    - specifications : Spec class for optimization problem specifications.
    - parameters : dict for optimization inputs names and fixed values.
    - wrapper : class to do the link between the specifications and the model.
    """
    def __init__(self, model: Callable[..., Dict], specifications: Spec,
                 parameters: Dict = {}, resultsHandler=True):
        self.specifications = specifications
        self.wrapper = Wrapper(model,specifications, parameters,resultsHandler)

    def run(self, ftol=1e-5, disp=True, maxiter=500,nbParetoPts=5,popsize=10,
            method='SLSQP',method2d='epsconstr',multi_start=0):
        """
        Solves the optimization problem.

        :param ftol: float: Precision goal for the value of f in the stopping criterion.
        :param disp: bool: Set to True to print convergence messages. If False, verbosity is ignored and set to 0.
        :param maxiter: int: maximum number of iterations.
        :param nbParetoPts: int: number of desired Pareto points.
        :param popsize: int: size of the population for the differential evolution algorithm (will be multiplied by number of optimization inputs).
        :param method: str: algorithm used to solve mono-objective problems.
        :param method2d: str: algorithm used to solve bi objective problems.
        :param multi_start: int : number of times the optimization is run with different random initial points
        :return: Wrapper class: results of optimization
        """
        # utilisation de SLSQP
        options = {'ftol': ftol, 'disp': disp, 'maxiter': maxiter,}
        pbSize = len(self.wrapper.spec.objectives)

        # TODO si pas de contraintes, lancer un BFGS

        if pbSize == 1:  # Cas mono-objectif
            if (method == 'SLSQP'):
                if multi_start==0:
                    result = SLSQP(self.wrapper, self.specifications.xinit,
                            options,disp,multi_start)
                else:
                    for i in range(multi_start):
                        x0 = numpy.random.rand(len(self.wrapper.spec.iNames))
                        print('initial point: ',denormalize(x0,self.wrapper.
                                                                spec.bounds))
                        x0 = jnp.array(x0)
                        result = SLSQP(self.wrapper,x0,options,disp,multi_start)
                        self.wrapper.printResults()

            elif (method=='IPOPT'):
                if multi_start==0:
                    result = IPOPT(self.wrapper,self.specifications.xinit,
                                   options,multi_start)
                else:
                    for i in range(multi_start):
                        x0 = numpy.random.rand(len(self.wrapper.spec.iNames))
                        print('initial point: ',denormalize(x0,self.wrapper.
                                                                spec.bounds))
                        x0 = jnp.array(x0)
                        result = IPOPT(self.wrapper,x0,options,multi_start)
                        self.wrapper.printResults()
            elif (method == 'LeastSquare'):
                result = LSSQ(self.wrapper, self.specifications.xinit, options)
                # TODO : modifier results
            elif (method == 'diffev'):
                result=diffev(self.wrapper,popsize,options)
            else:
                print("SLSQP or LeastSquare or IPOPT or diffev")

        elif pbSize == 2:  # Cas bi-objectif
            p = EpsilonConstraint(self.wrapper,method,disp)
            if method2d=='epsconstr':
                result = p.optim2D(self.specifications.xinit, nbParetoPts,
                                   options)
            elif method2d=='ponderation':
                result = p.optim2D_weighting(self.specifications.xinit,
                                             nbParetoPts, options)
            elif method2d=='sample':
                result = p.optim2D_basic(self.specifications.xinit,
                                         nbParetoPts, options)
            else:
                print('ponderation or epsconstr or sample')
            # TODO : modifier results
            # result = p.optim2D_basic(self.specifications.xinit, nbParetoPts,
            # options)
        else:
            print("Only mono or bi objective function.")

        return self.wrapper  # result  #TODO : modifier results


def SLSQP(wrapper: Wrapper, x0, options,disp,multi_start):
    """
    Solves the optimization problem with SLSQP method from scipy.optimize.minimize.

    :param wrapper: Class giving information to scipy.minimize (fgrad, bounds, constraints)
    :param x0: Initial inputs vector
    :param options: options = {'func': None, 'maxiter': 100, 'ftol': 1e-06,'iprint': 1, 'disp': False, 'eps': 1.4901161193847656e-08}
    with :
     - ftol : float Precision goal for the value of f in the stopping criterion.
     - eps : float  Step size used for numerical approximation of the Jacobian.
     - disp : bool  Set to True to print convergence messages. If False, verbosity is ignored and set to 0.
     - maxiter : int  Maximum number of iterations.
    :param disp: set to True to print convergence messages.
    :param multi_start: int : number of times the optimization is run with different random initial points
    :return: class Result including the optimal inputs vector
    """
    # Equality constraint means that the constraint function result is to
    # be zero
    # Inequality means that it is to be non-negative.
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
    wrapper.init()  # init doit être appelé avant l'optim pour bien prendre en
    # compte les nouvelles contraintes.
    options['disp']=False
    if multi_start==0:
        x0=normalize(x0,wrapper.spec.bounds) # normalization of values
    bounds=normalize_bounds(wrapper.spec.bounds)
    result = scipyOpt.minimize(wrapper.f_val_grad, x0=x0, jac=True,
                               method='SLSQP', bounds=bounds,
                               constraints=wrapper.constraints, options=options)
    while (not result.success) and (result.status!=9):
        if (result.status != 9): # not if maxiteration# TODO gérer une exception
            #  Iteration limit exceeded    (Exit mode 9)
            # TRY Random initial point !
            print("WARNING : Optimization doesn't converge... "
                  "Trying random inital guess")
            x0 = numpy.random.rand(len(x0))
            print('new initial point: ',denormalize(x0,wrapper.spec.bounds))
            x0=jnp.array(x0)
            result = scipyOpt.minimize(wrapper.f_val_grad, x0=x0, jac=True,
                            method='SLSQP', bounds=bounds,
                            constraints=wrapper.constraints, options=options)
    if (not result.success):
        print("ERROR : Optimization doesn't converge.")  # TODO donner plus
        # de détails sur les spec de cette optim
        print(result.message)
        # TODO faire afficher les specifications
        print('objectif : ' + str(wrapper.spec.objectives))
        print('contraintes : ' + str(wrapper.spec.eq_cstr)+
              str(wrapper.spec.ineq_cstr))
    result.fun=result.fun*(wrapper.spec.objectives_val[1]-
        wrapper.spec.objectives_val[0])+wrapper.spec.objectives_val[0]
    if (result.success) and disp:
        print(result.message,'\t(Exit mode 0)')
        print('\t\t\tCurrent function value:',result.fun)
        print('\t\t\tIterations:',result.nit)
        print('\t\t\tFunction evaluations:', result.nfev)
        print('\t\t\tGradient evaluations:', result.njev)
    result.x = denormalize(result.x, wrapper.spec.bounds)
    return result

def IPOPT(wrapper: Wrapper, x0,options,multi_start):
    """
    Solves the optimization problem with IPOPT method from scipy.optimize.minimize.

    :param wrapper: Class giving information to scipy.minimize (fgrad, bounds,constraints)
    :param x0: Initial inputs vector
    :param options: options = {'func': None, 'maxiter': 100, 'ftol': 1e-06,'iprint': 1, 'disp': False, 'eps': 1.4901161193847656e-08}
    with :
     - ftol : float Precision goal for the value of f in the stopping criterion.
     - eps : float  Step size used for numerical approximation of the Jacobian.
     - disp : bool  Set to True to print convergence messages. If False,verbosity is ignored and set to 0.
     - maxiter : int  Maximum number of iterations.
    :param multi_start: int : number of times the optimization is run with different random initial points
    :return: class Result including the optimal inputs vector
    """
    # Equality constraint means that the constraint function result is to
    # be zero
    # Inequality means that it is to be non-negative.
    # https://cyipopt.readthedocs.io/en/stable/tutorial.html
    wrapper.init()  # init doit être appelé avant l'optim pour bien prendre en
    # compte les nouvelles contraintes.
    if multi_start==0:
        x0=normalize(x0,wrapper.spec.bounds) # normalization of values
    bounds=normalize_bounds(wrapper.spec.bounds)
    from cyipopt import minimize_ipopt
    result = minimize_ipopt(wrapper.f_val_grad, x0=x0, jac=True,
                            bounds=bounds,constraints=wrapper.constraints)
    while (not result.success) and (result.status!=9):
        if (result.status != 9): # not if maxiteration# TODO gérer une exception
            #  Iteration limit exceeded    (Exit mode 9)
            # TRY Random initial point !
            print("WARNING : Optimization doesn't converge... "
                  "Trying random inital guess")
            x0 = numpy.random.rand(len(x0))
            print('new initial point: ',denormalize(x0,wrapper.spec.bounds))
            x0=jnp.array(x0)
            result = minimize_ipopt(wrapper.f_val_grad, x0=x0, jac=True,
                             bounds=bounds,constraints=wrapper.constraints)
    if (not result.success):
        print("ERROR : Optimization doesn't converge.")  # TODO donner plus
        # de détails sur les spec de cette optim
        print(result.message)
        # TODO faire afficher les specifications
        print('objectif : ' + str(wrapper.spec.objectives))
        print('contraintes : ' + str(wrapper.spec.eq_cstr)+
              str(wrapper.spec.ineq_cstr))
    result.fun=result.fun*(wrapper.spec.objectives_val[1]-
        wrapper.spec.objectives_val[0])+wrapper.spec.objectives_val[0]
    if (result.success)and options['disp']:
        print('Algorithm IPOPT terminated successfully')
        print('\t\t\tCurrent function value:',result.fun)
        print('\t\t\tIterations:',result.nit)
        print('\t\t\tFunction evaluations:', result.nfev)
        print('\t\t\tGradient evaluations:', result.njev)
    result.x=denormalize(result.x,wrapper.spec.bounds)
    return result

def diffev(wrapper: Wrapper, popsize, options):
    """
    Solves the optimization problem with differential_evolution stochastic method from scipy.optimize.minimize.

    :param wrapper: Class giving information to scipy.minimize (fgrad, bounds,constraints)
    :param popsize: int : Size of the population for the stochastic algorithm (will be multiplied by number of optimization inputs).
    :param options: options = {'func': None, 'maxiter': 100, 'ftol': 1e-06,'iprint': 1, 'disp': False, 'eps': 1.4901161193847656e-08}
    with:
     - ftol : float Precision goal for the value of f in the stopping criterion.
     - eps : float  Step size used for numerical approximation of the Jacobian.
     - disp : bool  Set to True to print convergence messages. If False, verbosity is ignored and set to 0.
     - maxiter : int  Maximum number of iterations.
    :return: class Result including the optimal inputs vector
    """
    wrapper.init()
    bounds=normalize_bounds(wrapper.spec.bounds)

    result = scipyOpt.differential_evolution(wrapper.f_penalty, bounds=bounds,
        args=(),strategy='best1bin',maxiter=options['maxiter'],popsize=popsize,
        atol=1e-15,tol=options['ftol'],mutation=(0.5,1),recombination=0.7,
        seed=None, callback=None,disp=False, polish=True,init='latinhypercube')

    if (not result.success):
        print("ERROR : Optimization doesn't converge.")
        print(result.message)
        print('objectif : ' + str(wrapper.spec.objectives))
        print('contraintes : ' + str(wrapper.spec.eq_cstr)+
              str(wrapper.spec.ineq_cstr))
    result.fun=result.fun*(wrapper.spec.objectives_val[1]-
        wrapper.spec.objectives_val[0])+wrapper.spec.objectives_val[0]
    if (result.success):
        print(result.message,'\t(Exit mode 0)')
        print('\t\t\tCurrent function value:',result.fun)
        print('\t\t\tIterations:',result.nit)
        print('\t\t\tFunction evaluations:', result.nfev)
    result.x=denormalize(result.x,wrapper.spec.bounds)
    return result

def LSSQ(wrapper: Wrapper, x0, options):
    """
    Solves the optimization problem with LeastSquare method from scipy.optimize.minimize.

    :param wrapper: Class giving information to scipy.minimize (fgrad, bounds, constraints)
    :param x0: Initial inputs vector
    :param options: options = {'func': None, 'maxiter': 100, 'ftol': 1e-06,'iprint': 1, 'disp': False, 'eps': 1.4901161193847656e-08}
    with :
     - ftol : float Precision goal for the value of f in the stopping criterion.
     - eps : float  Step size used for numerical approximation of the Jacobian.
     - disp : bool  Set to True to print convergence messages. If False, verbosity is ignored and set to 0.
     - maxiter : int  Maximum number of iterations.
    :return: class Result including the optimal inputs vector
    """
    bounds = np.array(normalize_bounds(wrapper.spec.bounds)).T
    x0 = normalize(x0, wrapper.spec.bounds)

    # https: // docs.scipy.org / doc / scipy / reference / generated /
    # scipy.optimize.least_squares.html
    result = scipyOpt.least_squares(fun=wrapper.f_val, x0=x0, bounds=bounds,
                                    jac=wrapper.f_grad,
                             ftol=options['ftol'], max_nfev=options['maxiter'])
    result.x=denormalize(result.x,wrapper.spec.bounds)
    result.fun=result.fun*(wrapper.spec.objectives_val[1]-
        wrapper.spec.objectives_val[0])+wrapper.spec.objectives_val[0]
    result.grad=result.grad*(wrapper.spec.objectives_val[1]-
        wrapper.spec.objectives_val[0])+wrapper.spec.objectives_val[0]
    if (not result.success or options['disp']):  # TODO gérer une exception
        print(result.message)
        print("Solution found: ", wrapper.getLastInputs())
        print("Value of the cost function at the solution: ", result.cost)
        print("Vector of residuals at the solution: ", result.fun)
        print("Gradient of the cost function at the solution: ", result.grad)

    return result

    '''
    is able to make several optimization changing one input parameter.
    returns the value of each elements of 'outputs' obtained for each optimization
    use iter.print() or noloadj.gui.plotIterations.plotXY(iter) to see the results
    '''


def optimizeParam(model: Callable[..., Dict], parameters: Dict,
                  specifications: Spec, variable, range, outputs,
                  ftol=0.001, disp=False, maxiter=50):
    """
    Solves all optimization problems for an input varying in a range of values,
    the other inputs remaining constant.

    :param model: model
    :param parameters: constant input parameters
    :param specifications: specifications of the problem
    :param variable: the varying input
    :param range: range of values of the varying input
    :param outputs: names of the outputs
    :param ftol: float Precision goal for the value of f in the stopping criterion.
    :param disp: bool Set to True to print convergence messages. If False, verbosity is ignored and set to 0.
    :param maxiter: int Maximum number of iterations.
    :return: list including all output vectors obtained at each optimization
    """
    iter = Iterations([variable], outputs)  # permet de sauvegarder les
    # résultats au fur et à mesure
    x = specifications.xinit
    for val in range:
        parameters[variable] = val
        optim = OptimProblem(model, specifications, parameters)
        res = optim.run(ftol=ftol, disp=disp, maxiter=maxiter)
        out = [res.rawResults[var] for var in outputs]
        iter.updateData([val], out)
    return iter



import jax.numpy as np
from jax.lax import *
from jax import custom_jvp,jvp
from functools import partial
from jax import config
config.update("jax_enable_x64", True)


def odeint45(f,x0,vect_t,*P,M=None,T=0.,h0=1e-5,tol=1.48e-8):
    '''
    Solves an ODE system described by f with Runge-Kutta 5 (7M) algorithm, and
    eventually compute its fast Fourier Transform (FFT).

    :param f: a class inherited from ODESystem abstract class that describes the ODE system
    :param x0: Array: initial state vector
    :param vect_t: Array: time vector on which the ODE system must be solved
    :param P: tuple of Array: optimization inputs
    :param M: int: number of points to compute the FFT
    :param T: float: operating period of the system (optional)
    :param h0: float: initial step size
    :param tol: float: tolerance for optimal step size computation

    :return: Several outputs:

    - Array : state vector 'xf' values across time vector
    - Array : output vector 'yf' values across time vector
    - int/Array : configurations 'states' across time vector is the system is periodic.
    '''
    return _odeint45(f,h0,tol,M,x0,vect_t,T,*P)


def rk_step(x_prev, t_prev, h_prev,f,*P):
    '''
    An iteration of the Runge-Kutta 5 (7M) algorithm.
    An iteration of the Runge-Kutta 5 (7M) algorithm.

    :param x_prev: the state vector at the previous iteration
    :param t_prev: the time at the previous iteration
    :param h_prev: step size at the previous iteration
    :param f: a class that describes the ODE system
    :param P: optimization inputs
    :return: the state vector at the present iteration x_now, its estimation
        x_now_est the time t_now at the present iteration.
    '''
    k1=f.timederivatives(x_prev, t_prev,*P)
    k2 = f.timederivatives(x_prev + h_prev*0.2 * k1, t_prev + 0.2 * h_prev,*P)
    k3 = f.timederivatives(x_prev+h_prev*(3 * k1+9*k2)/40,t_prev+3*h_prev/10,*P)
    k4 = f.timederivatives(x_prev + h_prev*(44 *k1/45-56*k2/15+32*k3/9),t_prev +
           4 * h_prev / 5,*P)
    k5 = f.timederivatives(x_prev + h_prev*(19372 * k1/6561-25360 * k2 / 2187 +
            64448 * k3 / 6561- 212 * k4 / 729),
           t_prev + 8 * h_prev / 9,*P)
    k6 = f.timederivatives(x_prev +h_prev*(9017*k1/3168-355*k2/33+46732*k3/5247+
            49 * k4 / 176 - 5103 * k5 / 18656),t_prev + h_prev,*P)
    k7 = f.timederivatives(x_prev + h_prev*(35 * k1 / 384 + 500 * k3 / 1113 +
            125 * k4 / 192 -2187 * k5 / 6784 + 11 * k6 / 84),t_prev + h_prev,*P)

    x_now = x_prev + h_prev *(35 * k1 / 384 + 500 * k3 / 1113 + 125 * k4 / 192
             -2187 * k5 / 6784 + 11 * k6 / 84)
    x_now_est = x_prev + h_prev *(5179 * k1 / 57600 + 7571* k3/16695+393*k4/640
            - 92097 * k5 / 339200 + 187 * k6 / 2100 + k7 / 40)
    t_now = t_prev + h_prev
    return x_now, x_now_est, t_now


def optimal_step(x_now,x_now_est,h_prev,tol,x_prev):
    '''
    Computes the optimal step size for the next iteration.

    :param x_now: state vector at the present iteration
    :param x_now_est: estimation of the state vector x_now
    :param h_prev: step size from the present iteration
    :param tol: tolerance for optimal step size computation
    :param x_prev: state vector at the previous iteration
    :return: optimal step size for the next iteration
    '''
    e=np.sqrt(np.sum(((x_now-x_now_est)/
                      (tol+tol*np.maximum(np.abs(x_now),np.abs(x_prev))))**2))
    hopt=h_prev*(e)**(-1/5)
    h_now=np.minimum(1.5*h_prev,0.9*hopt)
    return h_now

def interpolation(x_prev,y_prev,t_prev,x_now,y_now,t_now,t_new):
    '''
    Computes state and output vectors at time t_new by linear interpolation.

    :param x_prev: state vector at the previous iteration
    :param y_prev: output vector at the previous iteration
    :param t_prev: time at the previous iteration
    :param x_now: state vector at the present iteration
    :param y_now: output vector at the present iteration
    :param t_now: time at the present iteration
    :param t_new: new time for present iteration
    :return: state and output vectors x_new and y_new at t_new.
    '''
    x_new=((x_prev-x_now)*t_new+(t_prev*x_now-t_now*x_prev))/(t_prev-t_now)
    y_new=((y_prev-y_now)*t_new+(t_prev*y_now-t_now*y_prev))/(t_prev-t_now)
    return x_new,y_new

def compute_new_point(x_now,x_prev,t_prev,t_now,y_prev,y_now,
                      diff_var_threshold_value,diff_var_threshold_value_prev):
    '''
    Computes new time t_new and vectors x_new, y_new for which a state variable
    equals a threshold value.

    :param x_now: state vector at the present iteration
    :param x_prev: state vector at the previous iteration
    :param t_prev: time at the previous iteration
    :param t_now: time at the present iteration
    :param y_prev: output vector at the previous iteration
    :param y_now: output vector at the present iteration
    :param diff_var_threshold_value: differences between the output variable and
        the threshold value at the present iteration
    :param diff_var_threshold_value_prev: differences between the output variable
        and the threshold value at the previous iteration
    :return: new time t_new and vectors x_new, y_new
    '''
    t_new=(-t_prev*diff_var_threshold_value+t_now*diff_var_threshold_value_prev)/\
          (diff_var_threshold_value_prev-diff_var_threshold_value)
    h_new=t_new-t_prev
    x_new,y_new=interpolation(x_prev,y_prev,t_prev,x_now,y_now,t_now,t_new)
    return x_new,h_new,t_new,y_new

def interp_state_chgt(x_prev,y_prev,y_target,t_prev,x_now,y_now,t_now,f,i_prev,
                      h_prev,c_now):
    '''
    Tests if there is a changement of configuration of the ODE system at the
    present time.

    :param x_prev: state vector at the previous iteration
    :param y_prev: output vector at the previous iteration
    :param y_target: new output vector after changement of configuration
    :param t_prev: time at the previous iteration
    :param x_now: state vector at the present iteration
    :param y_now: output vector at the present iteration
    :param t_now: time at the present iteration
    :param f: a class that describes the ODE system
    :param i_prev: configuration of the system at the previous iteration
    :param h_prev: step size from the present iteration
    :param c_now: command of the ODE system for the present iteration
    :return: the new state vector x_now, outputs vector y_now, time t_now,
        the optimal step size for next iteration h_now.
    '''
    ind=np.where(np.array_equiv(y_now,y_target),-1,np.argmax(np.abs(y_now-
                                                                    y_target)))
    threshold_value=y_target[ind]
    diff_var_threshold_value=y_now[ind]-threshold_value
    diff_var_threshold_value_prev=y_prev[ind]-threshold_value
    condition = np.bitwise_and(np.sign(diff_var_threshold_value)!=np.sign(
            diff_var_threshold_value_prev),np.not_equal(ind,-1))
    condition=np.bitwise_and(condition,np.bitwise_not(np.allclose(
            diff_var_threshold_value_prev,0.)))
    x_now,h_now,t_now,y_now=cond(condition,lambda state:compute_new_point(
        x_now,x_prev,t_prev,t_now,y_prev,y_now,diff_var_threshold_value,
        diff_var_threshold_value_prev),lambda state:state,(x_now,h_prev,t_now,
                                                           y_now))
    if not isinstance(f.commande,type(None)) and not isinstance(f.state,
                                                             type(None)):
        _,x_now,y_now=f.update(x_now,y_now,t_now,c_now)
    elif isinstance(f.commande,type(None)) and not isinstance(f.state,
                                                               type(None)):
        _,x_now,y_now=f.update(x_now,y_now,t_now)
    else:
        x_now,y_now=f.update(x_now,y_now,t_now)
    return x_now,y_now,h_now,t_now

def next_step_simulation(x_prev,t_prev,y_prev,i_prev,c_prev,h_prev,f,tol,T,*P):
    '''
    Computes the state vector, output vector and time at the present iteration.

    :param x_prev: the state vector at the previous iteration
    :param t_prev: the time at the previous iteration
    :param y_prev: the output at the previous iteration
    :param i_prev: configuration of the system at the previous iteration
    :param c_prev: command of the system at the previous iteration
    :param h_prev: step size from the present iteration
    :param f: a class that describes the ODE system
    :param tol: tolerance for optimal step size computation
    :param T: operating period of the system (facultative)
    :param P: optimization inputs
    :return: the state vector x_now, outputs vector y_now, time t_now,
        configuration i_now, commande c_now at the present iteration,
        the optimal step size for next iteration h_now.
    '''
    if not isinstance(f.state,type(None)):
        f.state=i_prev
    if 'initialize' in f.__class__.__dict__:
        P=f.initialize(*P)
    x_now,x_now_est,t_now=rk_step(x_prev,t_prev,h_prev,f,*P)
    y_now=f.output(x_now,t_now,*P)
    y_now_est=f.output(x_now_est,t_now,*P)
    if isinstance(y_now,tuple):
        x_now,y_now=y_now
        x_now_est=y_now_est[0]
    h_now,c_now,tpdi=0.,0,0.
    if 'update' in f.__class__.__dict__ and not isinstance(f.state,type(None)):
        if not isinstance(f.commande,type(None)):
            tpdi=f.commande.next_pdi(t_now)
            c_now=f.commande.next_command(t_now)
            i_now,_,y_target=f.update(x_now,y_now,t_now,c_now)
        else:
            c_now=c_prev
            i_now,_,y_target=f.update(x_now,y_now,t_now)
        x_now,y_now,h_now,t_now=cond(np.bitwise_not(np.allclose(i_prev,i_now)),
            lambda state:interp_state_chgt(x_prev,y_prev,y_target,t_prev,x_now,
            y_now,t_now,f,i_prev,h_prev,c_now), lambda state:state,(x_now,y_now,
            h_now,t_now))

        y_now=f.output(x_now,t_now,*P)
        if len(y_now)==2:
            x_now,y_now=y_now
        h_now= optimal_step(x_now,x_now_est, h_prev, tol,x_prev)
        if not isinstance(f.commande,type(None)):
            h_now=np.minimum(tpdi-t_now,h_now)
        h_now=np.where(np.bitwise_not(np.allclose(i_prev,i_now)),1e-9,h_now) # pour accelerer code
    else:
        i_now=i_prev
        c_now=c_now
    if 'update' in f.__class__.__dict__ and isinstance(f.state,type(None)):
        _,y_target=f.update(x_now,y_now,t_now)
        x_now,y_now,h_now,t_now=interp_state_chgt(x_prev,y_prev,y_target,t_prev,
                                    x_now,y_now,t_now,f,i_prev,h_prev,c_now)
        y_now=f.output(x_now,t_now,*P)
        h_now = optimal_step(x_now, x_now_est, h_prev, tol,x_prev)
    elif not not isinstance(f.state,type(None)) and not 'update' in \
                                                        f.__class__.__dict__:
        h_now = optimal_step(x_now, x_now_est, h_prev, tol,x_prev)
    return x_now,t_now,y_now,h_now,i_now,c_now

def compute_fft(xf,yf,M):
    '''
    Compute the FFT of state variables xf and output variables yf.

    :param xf:  a matrix xf with state vector values across time
    :param yf: a matrix yf with output vector values across time
    :param M: number of points to compute the FFT
    :return: the respective FFT module and phase of xf and yf
    '''
    xfft=np.fft.fft(xf,M)*2/M # fft de x avec normalisation
    xfft=xfft.at[:,0].divide(2)
    yfft=np.fft.fft(yf,M)*2/M # fft de y avec normalisation
    yfft=yfft.at[:,0].divide(2)
    moduleX,phaseX=np.abs(xfft),np.angle(xfft) # amplitude et phase de la fft de x
    moduleY,phaseY=np.abs(yfft),np.angle(yfft) # amplitude et phase de la fft de y

    return xf,moduleX[:,0:M//2],phaseX[:,0:M//2],yf,moduleY[:,0:M//2],\
           phaseY[:,0:M//2]# on retire les frequences negatives


@partial(custom_jvp,nondiff_argnums=(0,1,2,3))
def _odeint45(f,h0,tol,M,x0,vect_t,T,*P):

    def scan_fun(state,te):

        def cond_fn(state):
            x_prev,t_prev,y_prev,h_prev,i_prev,c_prev=state
            return (t_prev<te) & (h_prev>0) # condition d'arrêt de simulation

        def body_fn(state):
            x_prev,t_prev,y_prev,h_prev,i_prev,c_prev=state

            x_now,t_now,y_now,h_now,i_now,c_now=next_step_simulation(x_prev,
                                t_prev,y_prev,i_prev,c_prev,h_prev,f,tol,T,*P) # prochaine itération de simulation

            # to reach te
            h_now=np.minimum(h_now,te-t_prev)

            return x_now,t_now,y_now,h_now,i_now,c_now

        x_now,t_now,y_now,h_now,i_now,c_now = while_loop(cond_fn,body_fn,state)

        return (x_now,t_now,y_now,h_now,i_now,c_now),(x_now,y_now,i_now)

    if not isinstance(f.state,type(None)):
        i0=f.state
    else:
        i0=0
    if not isinstance(f.commande,type(None)):
        c0=f.commande.next_command(vect_t[0])
    else:
        c0=0
    tempP=None
    if 'initialize' in f.__class__.__dict__:
        tempP=P
        P=f.initialize(*P)
    y0=f.output(x0,vect_t[0],*P)
    if isinstance(y0,tuple):
        x0,y0=y0
    if 'initialize' in f.__class__.__dict__:
        P=tempP
    vect,(xf,yf,states)=scan(scan_fun,(x0,vect_t[0],y0,h0,i0,c0),vect_t[1:])
    if not isinstance(f.state,type(None)): #  pour eviter erreurs d'execution
        f.state=vect[4]
    xf=np.transpose(np.concatenate((x0[None], xf))) # valeurs stockées du vecteur d'état
    yf=np.transpose(np.concatenate((y0[None], yf))) # valeurs stockées du vecteur de sortie
    if not isinstance(f.state,type(None)): # si le systeme comporte plusieurs configurations
        if isinstance(i0,int):
            states=np.insert(states,0,i0) # configurations écrites sous la forme d'un entier
        else:
            states=np.transpose(np.concatenate((i0[None], states))) # configurations écrites sous la forme d'un vecteur
        if M==None: # s'il n'y a pas de simulation fréquentielle à faire
            return xf,yf,states
        else: # s'il y a une simulation fréquentielle à faire
            xf, modX, phX, yf, modY, phY = compute_fft(xf, yf, M)
            return xf, modX, phX, yf, modY, phY, states
    else: # si le système ne comporte pas plusieurs configurations
        if M==None: # s'il n'y a pas de simulation fréquentielle à faire
            return xf,yf
        else: # s'il y a une simulation fréquentielle à faire
            xf, modX, phX, yf, modY, phY = compute_fft(xf, yf, M)
            return xf, modX, phX, yf, modY, phY


@_odeint45.defjvp
def _odeint45_jvp(f,h0,tol,M, primals, tangents):
    '''
    Solves an ODE system described by df/dP with Runge-Kutta 5 (7M) algorithm.

    :param f: a class that describes the ODE system
    :param h0: initial step size
    :param tol: tolerance for optimal step size computation
    :param M: number of points to compute the FFT
    :param primals: tuple including initial state vector, time vector, operating
        period and optimization inputs.
    :param tangents: tuple including differentials of initial state vector,
        operating period and optimization inputs.
    :return: a matrix xf with state vector values across time, and a matrix yf
        with output vector values across time, their respective derivatives
        w.r.t P (dxf and dyf), and a matrix states with configurations across
        time for periodic systems. Eventually the FFT module and phase,
        and their derivatives w.r.t optimization inputs.
    '''
    x0, vect_t,T, *P = primals
    dx0, _,_, *dP = tangents
    nPdP = len(P)

    res = odeint45_etendu(f, nPdP, h0, tol, M, x0, dx0, vect_t, T,*P, *dP)

    if not isinstance(f.state,type(None)): #si le systeme comporte plusieurs configurations
        if M==None: # s'il n'y a pas de simulation fréquentielle à faire
            xf,dxf,yf,dyf,states=res
            return (xf,yf,states),(dxf,dyf,states)
        else: # s'il y a une simulation fréquentielle à faire
            xf,dxf,modX,phX,dmodX,dphX,yf,dyf,modY,phY,dmodY,dphY,states=res
            return (xf,modX,phX,yf,modY,phY,states),(dxf,dmodX,dphX,dmodY,dphY,
                                                     states)
    else: # si le système ne comporte pas plusieurs configurations
        if M==None: # s'il n'y a pas de simulation fréquentielle à faire
            xf,dxf,yf,dyf=res
            return (xf,yf),(dxf,dyf)
        else: # s'il y a une simulation fréquentielle à faire
            xf,dxf,modX,phX,dmodX,dphX,yf,dyf,modY,phY,dmodY,dphY=res
            return (xf,modX,phX,yf,modY,phY),(dxf,dmodX,dphX,dyf,dmodY,dphY)

def f_grads(x,dx, t, f,nPdP, *P_and_dP):
    '''
    Computes the derivatives of state vector w.r.t P.

    :param x: state vector values
    :param dx: differential of state vector
    :param t: time at present iteration
    :param f: a class that describes the ODE system
    :param nPdP: length of optimization inputs
    :param P_and_dP: optimization inputs values and differentials
    :return: the derivatives of state vector w.r.t P.
    '''
    P, dP = P_and_dP[:nPdP], P_and_dP[nPdP:]
    res, dres = jvp(f.timederivatives, (x, t, *P), (dx,0., *dP))
    return dres

def rk45_step_der(x_prev, t_prev, dx_prev,h_prev,f,nPdP,*P_and_dP):
    '''
    An iteration of the Runge-Kutta 5 (7M) algorithm for the ODE system described
    by df/dP.

    :param x_prev: the state vector at the previous iteration
    :param t_prev: the time at the previous iteration
    :param dx_prev: the differential of state vector at the previous iteration
    :param h: constant step size
    :param h_prev: step size for the present iteration
    :param f: a class that describes the ODE system.
    :param nPdP: length of optimization inputs
    :param P_and_dP: optimization inputs values and differentials
    :return: dx/dp and x at the present iteration.
    '''
    P=P_and_dP[:nPdP]
    k1=f.timederivatives(x_prev, t_prev,*P)
    k2 = f.timederivatives(x_prev + h_prev*0.2 * k1, t_prev + 0.2 * h_prev,*P)
    k3=f.timederivatives(x_prev + h_prev*(3*k1 +9*k2)/40,t_prev+3 *h_prev/10,*P)
    k4 = f.timederivatives(x_prev + h_prev*(44 * k1 / 45 - 56 * k2/ 15+32*k3/9),
                      t_prev + 4 * h_prev / 5,*P)
    k5 = f.timederivatives(x_prev + h_prev*(19372*k1/6561 - 25360 * k2 / 2187 +
            64448 * k3 / 6561- 212 * k4 / 729),t_prev + 8 * h_prev / 9,*P)

    dk1 = f_grads(x_prev, dx_prev, t_prev,f,nPdP, *P_and_dP)
    dk2 = f_grads(x_prev+ h_prev*0.2 * k1, dx_prev + h_prev * 0.2 * dk1,t_prev +
                0.2 * h_prev ,f,nPdP, *P_and_dP)
    dk3 = f_grads(x_prev+ h_prev*(3 * k1 + 9 * k2) / 40,dx_prev+h_prev *(3 * dk1
                + 9 * dk2) / 40,t_prev+3 * h_prev / 10,f,nPdP, *P_and_dP)
    dk4 = f_grads(x_prev+ h_prev*(44 * k1 / 45 - 56 * k2 / 15 + 32 * k3 / 9),
                dx_prev +h_prev*(44 * dk1 / 45 - 56 * dk2 /15+32*dk3/9),
                t_prev + 4 * h_prev / 5,f,nPdP,*P_and_dP)
    dk5 = f_grads(x_prev+ h_prev*(19372 * k1 / 6561 - 25360 * k2 / 2187 +
            64448 * k3 / 6561- 212 * k4 / 729), dx_prev + h_prev *
        (19372 * dk1 / 6561 - 25360*dk2/2187+ 64448 * dk3 / 6561 - 212 * dk4
            / 729),t_prev + 8 * h_prev / 9,f,nPdP, *P_and_dP)
    dk6 = f_grads(x_prev+ h_prev*(9017 * k1 / 3168 - 355* k2 / 33+46732*k3/5247+
            49 * k4 / 176 - 5103 * k5 / 18656),dx_prev+h_prev*(9017 *
            dk1 / 3168 -355 *dk2/33 +46732*dk3/5247 + 49 * dk4 / 176 - 5103 *
            dk5 / 18656),t_prev + h_prev,f,nPdP,*P_and_dP)
    dx_now = dx_prev + h_prev *(35 * dk1 / 384 + 500 * dk3 / 1113 +
            125 * dk4 / 192 - 2187 * dk5 / 6784 + 11 * dk6 / 84)
    return dx_now

def next_der_step_simulation(x_prev,t_prev,dx_prev,x_now,t_now,h_prev,f,
                             nPdP,chgt_state,Mat,dMat,*P_and_dP):
    '''
    Computes the state vector, output vector, their derivatives and time at
    the present iteration.

    :param x_prev: the state vector at the previous iteration
    :param t_prev: the time at the previous iteration
    :param dx_prev: the differentials of the state vector at the previous iteration
    :param h: constant step size
    :param x_now: state vector at the present iteration
    :param t_now: time at the present iteration
    :param h_prev: step size for the present iteration
    :param f: a class that describes the ODE system
    :param nPdP: length of optimization inputs
    :param chgt_state: bool which indicates if system changed configuration
                       the previous iteration
    :param Mat: state matrices at the present iteration
    :param dMat: derivatives of state matrices at the present iteration
    :param P_and_dP: optimization inputs values and differentials
    :return: the derivatives of state and output vectors dx_now, dy_now, and
    state matrices Mat and their derivatives at the next iteration.
    '''
    P,dP = P_and_dP[:nPdP],P_and_dP[nPdP:]
    if 'initialize' in f.__class__.__dict__:
        P,dP = cond(chgt_state,lambda state:jvp(f.initialize, (*P,), (*dP,)),
                    lambda state:state,(Mat,dMat))
        P_and_dP=P+dP
        Mat,dMat=P,dP
        nPdP=len(P)
    dx_now=rk45_step_der(x_prev,t_prev,dx_prev,h_prev,f,nPdP,*P_and_dP)
    dx_now=np.where(np.abs(dx_now)>1e12,np.sign(dx_now)*1e12,dx_now) # pour eviter gradients trop grands
    dy_now = jvp(f.output, (x_now, t_now, *P), (dx_now, 0., *dP))[1]
    if isinstance(dy_now,tuple):
        dx_now,dy_now=dy_now
    return dx_now,dy_now,Mat,dMat

def compute_der_fft(xf,dxf,yf,dyf,M):
    '''
    Compute the FFT of state variables xf and output variables yf, and their
        derivatives w.r.t optimization inputs.

    :param xf:  a matrix xf with state vector values across time
    :param dxf: derivatives of xf w.r.t optimization inputs.
    :param yf: a matrix yf with output vector values across time
    :param dyf: derivatives of yf w.r.t optimization inputs.
    :param M: number of points to compute the FFT.
    :return: the respective FFT module and phase of xf and yf, and their derivatives
        w.r.t optimization inputs.
    '''
    xfft=np.fft.fft(xf,M)*2/M  # fft de x avec normalisation
    dxfft=np.fft.fft(dxf,M)*2/M
    xfft,dxfft=xfft.at[:,0].divide(2),dxfft.at[:,0].divide(2)
    yfft=np.fft.fft(yf,M)*2/M  # fft de x avec normalisation
    dyfft=np.fft.fft(dyf,M)*2/M
    yfft,dyfft=yfft.at[:,0].divide(2),dyfft.at[:,0].divide(2)
    moduleX,phaseX=np.abs(xfft),np.angle(xfft)  # amplitude et phase de la fft de x
    dmoduleX,dphaseX=(2*np.real(dxfft)*np.real(xfft)+2*np.imag(dxfft)*
       np.imag(xfft))/(2.*moduleX),(np.imag(dxfft)*np.real(xfft)-np.real(dxfft)*
                       np.imag(xfft))/(moduleX**2)
    moduleY,phaseY=np.abs(yfft),np.angle(yfft)  # amplitude et phase de la fft de x
    dmoduleY,dphaseY=(2*np.real(dyfft)*np.real(yfft)+2*np.imag(dyfft)*
       np.imag(yfft))/(2.*moduleY),(np.imag(dyfft)*np.real(yfft)-np.real(dyfft)*
                       np.imag(yfft))/(moduleY**2)

    return xf,dxf,moduleX[:,0:M//2],phaseX[:,0:M//2],dmoduleX[:,0:M//2],\
           dphaseX[:,0:M//2],yf,dyf,moduleY[:,0:M//2],phaseY[:,0:M//2],\
           dmoduleY[:,0:M//2],dphaseY[:,0:M//2]

def odeint45_etendu(f,nPdP,h0,tol,M,x0,dx0,vect_t,T,*P_and_dP):
    P,dP = P_and_dP[:nPdP],P_and_dP[nPdP:]
    dh=vect_t[1]-vect_t[0]

    def scan_fun(state, te):

        def cond_fn(state):
            x_prev,t_prev,y_prev,h_prev,i_prev2,i_prev,c_prev=state
            return (t_prev<te) & (h_prev>0) # condition d'arrêt de simulation

        def body_fn(state):
            x_prev,t_prev,y_prev,h_prev,i_prev2,i_prev,c_prev=state

            x_now,t_now,y_now,h_now,i_now,c_now=next_step_simulation(x_prev,
                                t_prev,y_prev,i_prev,c_prev,h_prev,f,tol,T,*P) # prochaine itération de simulation

            # to reach te
            h_now=np.minimum(h_now,te-t_prev)

            return x_now,t_now,y_now,h_now,i_prev,i_now,c_now

        x_prev,dx_prev,t_prev,y_prev,dy_prev,h_prev,i_prev,c_prev,Mat,dMat,\
            chgt_state=state
        x_now,t_now,y_now,h_now,i_prev,i_now,c_now = while_loop(
            cond_fn,body_fn, (x_prev,t_prev,y_prev,h_prev,i_prev,i_prev,c_prev))

        if not isinstance(f.state,type(None)):# pour eviter erreurs d'execution
            f.state=i_prev
        dx_now,dy_now,Mat,dMat=next_der_step_simulation(x_prev,t_prev,dx_prev,
                            x_now,t_now,dh,f,nPdP,chgt_state,Mat,dMat,*P_and_dP) # prochaine itération de calcul des gradients
        chgt_state=np.bitwise_not(np.array_equal(i_now,i_prev))

        return (x_now,dx_now,t_now,y_now,dy_now,h_now,i_now,c_now,Mat,dMat,
                chgt_state),(x_now,dx_now,y_now,dy_now,i_now)

    for element in f.__dict__.keys(): # pour eviter erreurs d'execution
        if hasattr(f.__dict__[element],'primal'):
            f.__dict__[element]=f.__dict__[element].primal
    if not isinstance(f.state,type(None)):
        i0=f.state
    else:
        i0=0
    if not isinstance(f.commande,type(None)):
        c0=f.commande.next_command(vect_t[0])
    else:
        c0=0
    tempP,tempdP,Mat,dMat=None,None,0,0
    if 'initialize' in f.__class__.__dict__:
        tempP,tempdP=P,dP
        P,dP=jvp(f.initialize,(*P,),(*dP,))
        Mat,dMat=P,dP
    y0,dy0=jvp(f.output,(x0,vect_t[0],*P),(dx0,0.,*dP))
    if isinstance(y0,tuple):
        x0,y0=y0
        dx0,dy0=dy0
    if 'initialize' in f.__class__.__dict__:
        P,dP=tempP,tempdP
    chgt_state=False
    vect,(xf,dxf,yf,dyf,states)=scan(scan_fun,(x0,dx0,vect_t[0],y0,dy0,h0,i0,
                                            c0,Mat,dMat,chgt_state),vect_t[1:]) # simulation dynamique
    if not isinstance(f.state,type(None)):# pour eviter erreurs d'execution
        f.state=vect[6]
    xf=np.transpose(np.concatenate((x0[None], xf))) # valeurs stockées du vecteur d'état
    yf=np.transpose(np.concatenate((y0[None], yf))) # valeurs stockées du vecteur de sortie
    dxf=np.transpose(np.concatenate((dx0[None], dxf))) # gradients du vecteur d'état
    dyf = np.transpose(np.concatenate((dy0[None], dyf))) # gradients du vecteur de sortie
    if not isinstance(f.state,type(None)): #si le systeme comporte plusieurs configurations
        if isinstance(i0,int):
            states=np.insert(states,0,i0) # configurations écrites sous la forme d'un entier
        else:
            states=np.transpose(np.concatenate((i0[None], states))) # configurations écrites sous la forme d'un vecteur
        if M==None: # s'il n'y a pas de simulation fréquentielle à faire
            return xf,dxf,yf,dyf,states
        else: # s'il y a une simulation fréquentielle à faire
            xf, dxf, modX, phX, dmodX,dphX, yf, dyf, modY, phY, dmodY, dphY=\
                compute_der_fft(xf,dxf,yf,dyf,M)
            return xf,dxf,modX,phX,dmodX,dphX,yf,dyf,modY,phY,dmodY,dphY,states
    else: # si le système ne comporte pas plusieurs configurations
        if M==None:# s'il n'y a pas de simulation fréquentielle à faire
            return xf,dxf,yf,dyf
        else:# s'il y a une simulation fréquentielle à faire
            xf, dxf, modX, phX, dmodX,dphX, yf, dyf, modY, phY, dmodY, dphY=\
                compute_der_fft(xf,dxf,yf,dyf,M)
            return xf,dxf,modX,phX,dmodX,dphX,yf,dyf,modY,phY,dmodY,dphY
